from . import element_handle


def mock_js_handle(js_handle, page) -> None:
    # MouseMocking
    def as_element_mocker():
        element = js_handle._as_element()
        element_handle.mock_element_handle(element, page)
        return element

    js_handle._as_element = js_handle.as_element
    js_handle.as_element = as_element_mocker



from functools import partial
from random import choice as rchoice
from sys import modules, stderr
from typing import Literal, Optional, Tuple, Union

import playfulrequests
from playfulrequests.headers import Headers
from playfulrequests.reqs import *
from playfulrequests.response import ProcessResponse

from .client import TLSClient
from .cookies import RequestsCookieJar
from .toolbelt import CaseInsensitiveDict


class TLSSession(TLSClient):
    """
    Session object that sends requests with TLS client.

    Args:
        browser (str): Browser to use [firefox, chrome]
        client_identifier (str): Identifier for the client
        os (Literal['win', 'mac', 'lin'], optional): OS to use in header [win, mac, lin]
        headers (dict, optional): Dictionary of HTTP headers to send with the request. Default is generated from `browser` and `os`.
        temp (bool, optional): Indicates if session is temporary. Defaults to False.
        verify (bool, optional): Verify the server's TLS certificate. Defaults to True.
        timeout (int, optional): Default timeout in seconds. Defaults to 30.
        ja3_string (str, optional): JA3 string. Defaults to None.
        h2_settings (dict, optional): HTTP/2 settings. Defaults to None.
        additional_decode (str, optional): Additional decode. Defaults to None.
        pseudo_header_order (list, optional): Pseudo header order. Defaults to None.
        priority_frames (list, optional): Priority frames. Defaults to None.
        header_order (list, optional): Header order. Defaults to None.
        force_http1 (bool, optional): Force HTTP/1. Defaults to False.
        catch_panics (bool, optional): Catch panics. Defaults to False.
        debug (bool, optional): Debug mode. Defaults to False.

    Methods:
        get(url, *, params=None, headers=None, cookies=None, allow_redirects=True, verify=None, timeout=30, proxies=None):
            Send a GET request
        post(url, *, params=None, data=None, files=None, headers=None, cookies=None, json=None, allow_redirects=True, verify=None, timeout=30, proxies=None):
            Send a POST request
        options(url, *, params=None, headers=None, cookies=None, allow_redirects=True, verify=None, timeout=30, proxies=None):
            Send a OPTIONS request
        head(url, *, params=None, headers=None, cookies=None, allow_redirects=True, verify=None, timeout=30, proxies=None):
            Send a HEAD request
        put(url, *, params=None, data=None, files=None, headers=None, cookies=None, json=None, allow_redirects=True, verify=None, timeout=30, proxies=None):
            Send a PUT request
        patch(url, *, params=None, data=None, files=None, headers=None, cookies=None, json=None, allow_redirects=True, verify=None, timeout=30, proxies=None):
            Send a PATCH request
        delete(url, *, params=None, headers=None, cookies=None, allow_redirects=True, verify=None, timeout=30, proxies=None):
            Send a DELETE request
        render(url, headless, proxy, response, mock_human):
            Render a page with playwright
    """

    def __init__(
        self,
        client_identifier: str,
        browser: Literal['firefox', 'chrome'],
        os: Optional[Literal['win', 'mac', 'lin']] = None,
        headers: Optional[dict] = None,
        temp: bool = False,
        verify: bool = True,
        timeout: float = 30,
        *args,
        **kwargs,
    ):
        super().__init__(client_identifier=client_identifier, *args, **kwargs)

        self.browser: str = browser  # browser name

        # sync network methods
        self.get: partial = partial(get, session=self)
        self.post: partial = partial(post, session=self)
        self.options: partial = partial(options, session=self)
        self.head: partial = partial(head, session=self)
        self.put: partial = partial(put, session=self)
        self.patch: partial = partial(patch, session=self)
        self.delete: partial = partial(delete, session=self)

        # async network methods
        self.async_get: partial = partial(async_get, session=self)
        self.async_post: partial = partial(async_post, session=self)
        self.async_options: partial = partial(async_options, session=self)
        self.async_head: partial = partial(async_head, session=self)
        self.async_put: partial = partial(async_put, session=self)
        self.async_patch: partial = partial(async_patch, session=self)
        self.async_delete: partial = partial(async_delete, session=self)

        self.temp: bool = temp  # indicate if session is temporary
        self._os: str = os or rchoice(('win', 'mac', 'lin'))  # os name
        self.verify: bool = verify  # default to verifying certs
        self.timeout: float = timeout  # default timeout

        # set headers
        if headers:
            self.headers = CaseInsensitiveDict(headers)
        else:
            self.resetHeaders(os=os)

    def resetHeaders(
        self,
        os: Optional[Literal['win', 'mac', 'lin']] = None,
    ):
        """
        Rotates the headers of the session
        "OS" can be one of the following:
          ['win', 'mac', 'lin']
        Default is what it was initialized with, or the last value set
        """
        if os:
            self._os = os
        self.headers = CaseInsensitiveDict(Headers(browser=self.browser, os=self._os).generate())

    @property
    def os(self) -> str:
        return self._os

    @os.setter
    def os(self, os: Literal['win', 'mac', 'lin']):
        if os not in _os_set:
            raise ValueError(f'`{os}` is not a valid OS: (win, mac, lin)')
        self.resetHeaders(os=os)

    def render(self, *args, **kwargs):
        # shortcut to render method
        if 'playwright' in modules:
            return playfulrequests.browser.render(*args, **kwargs, session=self, browser=self.browser)
        else:
            stderr.write('Cannot render. Playwright not installed.\n')

    def request(
        self,
        method: str,
        url: str,
        *,
        data: Optional[Union[str, bytes, bytearray, dict]] = None,
        files: Optional[dict] = None,
        headers: Optional[Union[dict, CaseInsensitiveDict]] = None,
        cookies: Optional[Union[RequestsCookieJar, dict, list]] = None,
        json: Optional[Union[dict, list, str]] = None,
        allow_redirects: bool = True,
        history: bool = False,
        verify: Optional[bool] = None,
        timeout: Optional[float] = None,
        proxies: Optional[dict] = None,
        process: bool = True,
    ) -> 'playfulrequests.response.Response':
        """
        Send a request with TLS client

        Args:
            method (str): Method of request (GET, POST, OPTIONS, HEAD, PUT, PATCH, DELETE)
            url (str): URL to send request to
            data (Union[str, bytes, bytearray, dict], optional): Data to send to request. Defaults to None.
            files (dict, optional): Files to send with request. Defaults to None.
            headers (dict, optional): Dictionary of HTTP headers to send with the request. Defaults to None.
            cookies (Union[RequestsCookieJar, dict, list], optional): Dict or CookieJar to send. Defaults to None.
            json (dict, optional): Json to send in the request body. Defaults to None.
            allow_redirects (bool, optional): Allow request to redirect. Defaults to True.
            history (bool, optional): Remember request history. Defaults to False.
            verify (bool, optional): Verify the server's TLS certificate. Defaults to True.
            timeout (float, optional): Timeout in seconds. Defaults to 30.
            proxies (dict, optional): Dictionary of proxies. Defaults to None.

        Returns:
            playfulrequest.response.Response: Response object
        """
        proc = ProcessResponse(
            session=self,
            method=method,
            url=url,
            data=data,
            files=files,
            headers=headers,
            cookies=cookies,
            json=json,
            allow_redirects=allow_redirects,
            history=history,
            verify=self.verify if verify is None else verify,
            timeout=self.timeout if timeout is None else timeout,
            proxy=proxies,
        )
        if not process:
            # return an unfinished ProcessResponse object
            return proc
        proc.send()
        return proc.response


class Session(TLSSession):
    def __init__(
        self,
        browser: Literal['firefox', 'chrome'] = 'chrome',
        version: Optional[int] = None,
        os: Optional[Literal['win', 'mac', 'lin']] = None,
        headers: Optional[dict] = None,
        *args,
        **kwargs,
    ):
        '''
        Parameters:
            browser (Literal['firefox', 'chrome'], optional): Browser to use. Default is 'firefox'.
            version (int, optional): Version of the browser to use. Browser must be specified. Default is randomized.
            os (Literal['win', 'mac', 'lin'], optional): OS to use in header. Default is randomized.
            headers (dict, optional): Dictionary of HTTP headers to send with the request. Default is generated from `browser` and `os`.
            verify (bool, optional): Verify the server's TLS certificate. Defaults to True.
            ja3_string (str, optional): JA3 string. Defaults to None.
            h2_settings (dict, optional): HTTP/2 settings. Defaults to None.
            additional_decode (str, optional): Additional decode. Defaults to None.
            pseudo_header_order (list, optional): Pseudo header order. Defaults to None.
            priority_frames (list, optional): Priority frames. Defaults to None.
            header_order (list, optional): Header order. Defaults to None.
            force_http1 (bool, optional): Force HTTP/1. Defaults to False.
            catch_panics (bool, optional): Catch panics. Defaults to False.
            debug (bool, optional): Debug mode. Defaults to False.
        '''
        # random version if not specified
        if not version:
            version = _browsers[browser].version
        # if version is specified, check if it is supported
        elif version not in _browsers[browser].versions:
            raise ValueError(
                f'`{version}` is not a supported {browser} version: {_browsers[browser].versions}'
            )
        self.version = version

        super().__init__(
            client_identifier=f'{browser}_{version}',
            browser=browser,
            headers=headers,
            os=os,
            *args,
            **kwargs,
        )


class SessionShortcut:
    name: str
    versions: Tuple[int]

    @classmethod
    @property
    def version(cls) -> int:
        return rchoice(cls.versions)

    @classmethod
    def Session(
        cls,
        version: Optional[int] = None,
        os: Optional[Literal['win', 'mac', 'lin']] = None,
        *args,
        **kwargs,
    ) -> Session:
        return Session(
            browser=cls.name,
            version=version or cls.version,
            os=os,
            *args,
            **kwargs,
        )

    @classmethod
    def BrowserSession(
        cls,
        **kwargs,
    ):
        return playfulrequests.browser.BrowserSession(
            browser=cls.name,
            **kwargs,
        )


class firefox(SessionShortcut):
    name: str = 'firefox'
    versions: Tuple[int] = (102, 104, 105, 106, 108, 110, 117)


class chrome(SessionShortcut):
    name: str = 'chrome'
    versions: Tuple[int] = (103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 117)


_browsers: dict = {'firefox': firefox, 'chrome': chrome}
_os_set: set = {'win', 'mac', 'lin'}



import typing

from . import element_handle, frame_locator, js_handle, locator


def mock_element_handle(element, page) -> None:
    async def click_mocker(
        button="left",
        click_count=1,
        delay=20,
        force=False,
        modifiers=[],
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await ElementHandle.click(
            element,
            page,
            button=button,
            click_count=click_count,
            delay=delay,
            force=force,
            modifiers=modifiers,
            no_wait_after=no_wait_after,
            position=position,
            timeout=timeout,
            trial=trial,
        )

    element.click = click_mocker

    async def dblclick_mocker(
        button="left",
        delay=20,
        force=False,
        modifiers=[],
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await ElementHandle.dblclick(
            element,
            page,
            button=button,
            delay=delay,
            force=force,
            modifiers=modifiers,
            no_wait_after=no_wait_after,
            position=position,
            timeout=timeout,
            trial=trial,
        )

    element.dblclick = dblclick_mocker

    async def check_mocker(
        force=False,
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await ElementHandle.check(
            element,
            page,
            force=force,
            no_wait_after=no_wait_after,
            position=position,
            timeout=timeout,
            trial=trial,
        )

    element.check = check_mocker

    async def uncheck_mocker(
        force=False,
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await ElementHandle.uncheck(
            element,
            page,
            force=force,
            no_wait_after=no_wait_after,
            position=position,
            timeout=timeout,
            trial=trial,
        )

    element.uncheck = uncheck_mocker

    async def set_checked_mocker(
        checked=False,
        force=False,
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await ElementHandle.set_checked(
            element,
            page,
            checked=checked,
            force=force,
            no_wait_after=no_wait_after,
            position=position,
            timeout=timeout,
            trial=trial,
        )

    element.set_checked = set_checked_mocker

    async def hover_mocker(
        force=False, modifiers=[], position={}, timeout: typing.Optional[float] = None, trial=False
    ):
        await ElementHandle.hover(
            element,
            page,
            force=force,
            modifiers=modifiers,
            position=position,
            timeout=timeout,
            trial=trial,
        )

    element.hover = hover_mocker

    async def type_mocker(
        text, delay=50, no_wait_after=False, timeout: typing.Optional[float] = None
    ):
        await ElementHandle.type(
            element, page, text, delay=delay, no_wait_after=no_wait_after, timeout=timeout
        )

    element.type = type_mocker

    # ElementHandle
    async def mock_query_selector(selector, strict=False) -> typing.Optional["ElementHandle"]:
        element = await element._query_selector(selector, strict=strict)
        if element:
            await element_handle.mock_element_handle(element, page)
        return element

    element._query_selector = element.query_selector
    element.query_selector = mock_query_selector

    async def mock_query_selector_all(selector) -> typing.List["ElementHandle"]:
        elements = await element._query_selector_all(selector)
        for element in elements:
            await element_handle.mock_element_handle(element, page)
        return elements

    element._query_selector_all = element.query_selector_all
    element.query_selector_all = mock_query_selector_all

    async def mock_wait_for_selector(
        selector, state=[], strict=False, timeout: typing.Optional[float] = None
    ) -> typing.Optional["ElementHandle"]:
        element = await element.__wait_for_selector(
            selector, state=state, strict=strict, timeout=timeout
        )
        if element:
            await element_handle.mock_element_handle(element, page)
        return element

    element.__wait_for_selector = element.wait_for_selector
    element.wait_for_selector = mock_wait_for_selector

    # JsHandle
    async def mock_evaluate_handle(expression, arg=None) -> "JSHandle":
        _js_handle = await element._evaluate_handle(expression, arg=arg)
        await js_handle.mock_js_handle(_js_handle, page)
        return _js_handle

    element._evaluate_handle = element.evaluate_handle
    element.evaluate_handle = mock_evaluate_handle


class ElementHandle:
    async def click(
        element,
        page,
        button="left",
        click_count=1,
        delay=20,
        force=False,
        modifiers=[],
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        frame = element.owner_frame()

        if not force:
            await element.wait_for_element_state("editable", timeout=timeout)

        if not trial:
            if page.scroll_into_view:
                await element.scroll_into_view_if_needed(timeout=timeout)

            boundings = await element.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            for modifier in modifiers:
                await frame.page.keyboard.down(modifier)

            await frame.page.mouse.click(x, y, button, click_count, delay)

            for modifier in modifiers:
                await frame.page.keyboard.up(modifier)

    async def dblclick(
        element,
        page,
        button="left",
        delay=20,
        force=False,
        modifiers=[],
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        frame = element.owner_frame()

        if not force:
            await element.wait_for_element_state("editable", timeout=timeout)

        if not trial:
            if page.scroll_into_view:
                await element.scroll_into_view_if_needed(timeout=timeout)

            boundings = await element.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            for modifier in modifiers:
                await frame.page.keyboard.down(modifier)

            await frame.page.mouse.dblclick(x, y, button, delay)

            for modifier in modifiers:
                await frame.page.keyboard.up(modifier)

    async def check(
        element,
        page,
        force=False,
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        frame = element.owner_frame()

        if not force:
            await element.wait_for_element_state("editable", timeout=timeout)

        if await element.is_checked():
            return

        if not trial:
            if page.scroll_into_view:
                await element.scroll_into_view_if_needed(timeout=timeout)

            boundings = await element.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            await frame.page.mouse.click(x, y, button="left", click_count=1, delay=20)

            assert await element.is_checked()

    async def uncheck(
        element,
        page,
        force=False,
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        frame = element.owner_frame()

        if not force:
            await element.wait_for_element_state("editable", timeout=timeout)

        if not await element.is_checked():
            return

        if not trial:
            if page.scroll_into_view:
                await element.scroll_into_view_if_needed(timeout=timeout)

            boundings = await element.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            await frame.page.mouse.click(x, y, button="left", click_count=1, delay=20)

            assert not await element.is_checked()

    async def set_checked(
        element,
        page,
        checked=False,
        force=False,
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        frame = element.owner_frame()

        if not force:
            await element.wait_for_element_state("editable", timeout=timeout)

        if await element.is_checked() == checked:
            return

        if not trial:
            if page.scroll_into_view:
                await element.scroll_into_view_if_needed(timeout=timeout)

            boundings = await element.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            await frame.page.mouse.click(x, y, button="left", click_count=1, delay=20)

            assert await element.is_checked()

    async def hover(
        element,
        page,
        force=False,
        modifiers=[],
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        frame = element.owner_frame()

        if not force:
            await element.wait_for_element_state("editable", timeout=timeout)

        if not trial:
            if page.scroll_into_view:
                await element.scroll_into_view_if_needed(timeout=timeout)

            boundings = await element.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            for modifier in modifiers:
                await frame.page.keyboard.down(modifier)

            await frame.page.mouse.move(x, y)

            for modifier in modifiers:
                await frame.page.keyboard.up(modifier)

    async def type(
        element, page, text, delay=50, no_wait_after=False, timeout: typing.Optional[float] = None
    ) -> None:
        frame = element.owner_frame()

        await element.wait_for_element_state("editable", timeout=timeout)

        if page.scroll_into_view:
            await element.scroll_into_view_if_needed(timeout=timeout)

        boundings = await element.bounding_box()
        x, y, width, height = boundings.values()

        x, y = x + width // 2, y + height // 2

        await frame.page.mouse.click(x, y, "left", 1, delay)

        await frame.page.keyboard.type(text, delay=delay)



from .response import Response, ProcessResponse
from .session import Session, TLSSession, chrome, firefox
from .reqs import *
from .headers import Headers

# attempt to import headless browsing dependencies
try:
    from .playwright_mock import ChromeBrowser, FirefoxBrowser
    from .browser import BrowserSession, render
except ModuleNotFoundError:
    import os
    from sys import stderr

    # give windows users a warning
    os.name == 'nt' and stderr.write(
        'WARNING: Please run `pip install playfulrequests[all]` for headless browsing support.\n'
    )

from .parser import HTML
from .__version__ import __version__
from .__version__ import __author__



import uuid
from dataclasses import dataclass
from typing import Dict, List, Optional, Union
from urllib.parse import urlencode

from geventhttpclient import HTTPClient
from orjson import dumps, loads

import playfulrequests
from playfulrequests.cffi import PORT, destroy_session

from .cookies import (
    RequestsCookieJar,
    cookiejar_from_dict,
    cookiejar_to_list,
    extract_cookies_to_jar,
    list_to_cookiejar,
    merge_cookies,
)
from .exceptions import ClientException
from .toolbelt import CaseInsensitiveDict

try:
    import turbob64 as base64
except ImportError:
    import base64

'''
TLSClient heavily based on https://github.com/FlorianREGAZ/Python-Tls-Client
Copyright (c) 2022 Florian Zager
'''


@dataclass
class TLSClient:
    client_identifier: Optional[str] = None
    random_tls_extension_order: bool = True
    force_http1: bool = False
    catch_panics: bool = False
    debug: bool = False
    proxies: Optional[dict] = None
    cookies: Optional[RequestsCookieJar] = None

    # custom TLS profile
    ja3_string: Optional[str] = None
    h2_settings: Optional[Dict[str, int]] = None
    h2_settings_order: Optional[List[str]] = None
    supported_signature_algorithms: Optional[List[str]] = None
    supported_delegated_credentials_algorithms: Optional[List[str]] = None
    supported_versions: Optional[List[str]] = None
    key_share_curves: Optional[List[str]] = None
    cert_compression_algo: Optional[str] = None
    additional_decode: Optional[str] = None
    pseudo_header_order: Optional[List[str]] = None
    connection_flow: Optional[int] = None
    priority_frames: Optional[list] = None
    header_order: Optional[List[str]] = None
    header_priority: Optional[List[str]] = None

    '''
    Synopsis:
    
    self.client_identifier examples:
    - Chrome > chrome_103, chrome_104, chrome_105, chrome_106
    - Firefox > firefox_102, firefox_104
    - Opera > opera_89, opera_90
    - Safari > safari_15_3, safari_15_6_1, safari_16_0
    - iOS > safari_ios_15_5, safari_ios_15_6, safari_ios_16_0
    - iPadOS > safari_ios_15_6

    self.ja3_string example:
    - 771,4865-4866-4867-49195-49199-49196-49200-52393-52392-49171-49172-156-157-47-53,0-23-65281-10-11-35-16-5-13-18-51-45-43-27-17513,29-23-24,0

    HTTP2 Header Frame Settings
    - HEADER_TABLE_SIZE
    - SETTINGS_ENABLE_PUSH
    - MAX_CONCURRENT_STREAMS
    - INITIAL_WINDOW_SIZE
    - MAX_FRAME_SIZE
    - MAX_HEADER_LIST_SIZE

    self.h2_settings example:
    {
        "HEADER_TABLE_SIZE": 65536,
        "MAX_CONCURRENT_STREAMS": 1000,
        "INITIAL_WINDOW_SIZE": 6291456,
        "MAX_HEADER_LIST_SIZE": 262144
    }

    HTTP2 Header Frame Settings Order
    self.h2_settings_order example:
    [
        "HEADER_TABLE_SIZE",
        "MAX_CONCURRENT_STREAMS",
        "INITIAL_WINDOW_SIZE",
        "MAX_HEADER_LIST_SIZE"
    ]

    Supported Signature Algorithms
    - PKCS1WithSHA256
    - PKCS1WithSHA384
    - PKCS1WithSHA512
    - PSSWithSHA256
    - PSSWithSHA384
    - PSSWithSHA512
    - ECDSAWithP256AndSHA256
    - ECDSAWithP384AndSHA384
    - ECDSAWithP521AndSHA512
    - PKCS1WithSHA1
    - ECDSAWithSHA1

    self.supported_signature_algorithms example:
    [
        "ECDSAWithP256AndSHA256",
        "PSSWithSHA256",
        "PKCS1WithSHA256",
        "ECDSAWithP384AndSHA384",
        "PSSWithSHA384",
        "PKCS1WithSHA384",
        "PSSWithSHA512",
        "PKCS1WithSHA512",
    ]

    Supported Delegated Credentials Algorithms
    - PKCS1WithSHA256
    - PKCS1WithSHA384
    - PKCS1WithSHA512
    - PSSWithSHA256
    - PSSWithSHA384
    - PSSWithSHA512
    - ECDSAWithP256AndSHA256
    - ECDSAWithP384AndSHA384
    - ECDSAWithP521AndSHA512
    - PKCS1WithSHA1
    - ECDSAWithSHA1

    self.supported_delegated_credentials_algorithms example:
    [
        "ECDSAWithP256AndSHA256",
        "PSSWithSHA256",
        "PKCS1WithSHA256",
        "ECDSAWithP384AndSHA384",
        "PSSWithSHA384",
        "PKCS1WithSHA384",
        "PSSWithSHA512",
        "PKCS1WithSHA512",
    ]

    Supported Versions
    - GREASE
    - 1.3
    - 1.2
    - 1.1
    - 1.0

    self.supported_versions example:
    [
        "GREASE",
        "1.3",
        "1.2"
    ]

    Key Share Curves
    - GREASE
    - P256
    - P384
    - P521
    - X25519

    self.key_share_curves example:
    [
        "GREASE",
        "X25519"
    ]

    Cert Compression Algorithm
    self.cert_compression_algo examples: "zlib", "brotli", "zstd"

    Additional Decode
    Make sure the go code decodes the response body once explicit by provided algorithm.
    self.additional_decode examples: null, "gzip", "br", "deflate"

    Pseudo Header Order (:authority, :method, :path, :scheme)
    self.pseudo_header_order examples:
    [
        ":method",
        ":authority",
        ":scheme",
        ":path"
    ]

    Connection Flow / Window Size Increment
    self.connection_flow example:
    15663105

    self.priority_frames example:
    [
        {
        "streamID": 3,
        "priorityParam": {
            "weight": 201,
            "streamDep": 0,
            "exclusive": false
        }
        },
        {
        "streamID": 5,
        "priorityParam": {
            "weight": 101,
            "streamDep": false,
            "exclusive": 0
        }
        }
    ]

    Order of your headers
    self.header_order example:
    [
        "key1",
        "key2"
    ]

    Header Priority
    self.header_priority example:
    {
        "streamDep": 1,
        "exclusive": true,
        "weight": 1
    }
    
    Proxies
    self.proxies usage:
    {
        "http": "http://user:pass@ip:port",
        "https": "http://user:pass@ip:port"
    }
    '''

    def __post_init__(self) -> None:
        self._session_id: str = str(uuid.uuid4())

        self.headers: CaseInsensitiveDict = CaseInsensitiveDict()
        self.proxies: dict = self.proxies or {}

        # http client for local go server
        self.server: HTTPClient = HTTPClient(
            '127.0.0.1', PORT, ssl=False, insecure=True, connection_timeout=1e9, network_timeout=1e9
        )
        # CookieJar containing all currently outstanding cookies set on this session
        self.cookies: RequestsCookieJar = self.cookies or RequestsCookieJar()
        self._closed: bool = False  # indicate if session is closed

    def close(self):
        if not self._closed:
            self._closed = True
            destroy_session(self._session_id)
            self.server.close()

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()

    def __del__(self):
        self.close()

    def build_request(
        self,
        method: str,
        url: str,
        headers: Optional[Union[dict, CaseInsensitiveDict]] = None,
        data: Optional[Union[str, bytes, bytearray, dict]] = None,
        cookies: Optional[Union[RequestsCookieJar, dict, list]] = None,
        json: Optional[Union[dict, list, str]] = None,
        allow_redirects: bool = False,
        history: bool = True,
        verify: Optional[bool] = None,
        timeout: Optional[float] = None,
        proxy: Optional[dict] = None,
    ):
        # Prepare request body - build request body
        # Data has priority. JSON is only used if data is None.
        if data is None and json is not None:
            if type(json) in (dict, list):
                json = dumps(json).decode('utf-8')
            request_body = json
            content_type = 'application/json'
        elif data is not None and type(data) not in (str, bytes):
            request_body = urlencode(data, doseq=True)
            content_type = 'application/x-www-form-urlencoded'
        else:
            request_body = data
            content_type = None
        # set content type if it isn't set
        if content_type is not None and 'content-type' not in self.headers:
            self.headers['Content-Type'] = content_type

        # Prepare Headers
        if self.headers is None:
            headers = CaseInsensitiveDict(headers)
        elif headers is None:
            headers = self.headers
        else:
            merged_headers = CaseInsensitiveDict(self.headers)
            merged_headers.update(headers)

            # Remove items, where the key or value is set to None.
            none_keys = [k for (k, v) in merged_headers.items() if v is None or k is None]
            for key in none_keys:
                del merged_headers[key]

            headers = merged_headers

        # Cookies
        if isinstance(cookies, RequestsCookieJar):
            merge_cookies(self.cookies, cookies)
        elif isinstance(cookies, list):
            merge_cookies(self.cookies, list_to_cookiejar(cookies))
        elif isinstance(cookies, dict):
            self.cookies.set_cookie(cookiejar_from_dict(cookies))

        # turn cookie jar into dict

        # Proxy
        proxy = proxy or self.proxies

        if type(proxy) is dict and 'http' in proxy:
            proxy = proxy['http']
        elif type(proxy) is str:
            proxy = proxy
        else:
            proxy = ''

        # Request
        is_byte_request = isinstance(request_body, (bytes, bytearray))
        request_payload = {
            'sessionId': self._session_id,
            'followRedirects': allow_redirects,
            'wantHistory': history,
            'forceHttp1': self.force_http1,
            'withDebug': self.debug,
            'catchPanics': self.catch_panics,
            'headers': dict(headers) if isinstance(headers, CaseInsensitiveDict) else headers,
            'headerOrder': self.header_order,
            'insecureSkipVerify': not verify,
            'isByteRequest': is_byte_request,
            'additionalDecode': self.additional_decode,
            'proxyUrl': proxy,
            'requestUrl': url,
            'requestMethod': method,
            'requestBody': base64.b64encode(request_body).decode()
            if is_byte_request
            else request_body,
            'requestCookies': cookiejar_to_list(self.cookies),
            'timeoutMilliseconds': int(timeout * 1000),
        }
        if self.client_identifier is None:
            request_payload['customTlsClient'] = {
                'ja3String': self.ja3_string,
                'h2Settings': self.h2_settings,
                'h2SettingsOrder': self.h2_settings_order,
                'pseudoHeaderOrder': self.pseudo_header_order,
                'connectionFlow': self.connection_flow,
                'priorityFrames': self.priority_frames,
                'headerPriority': self.header_priority,
                'certCompressionAlgo': self.cert_compression_algo,
                'supportedVersions': self.supported_versions,
                'supportedSignatureAlgorithms': self.supported_signature_algorithms,
                'supportedDelegatedCredentialsAlgorithms': self.supported_delegated_credentials_algorithms,
                'keyShareCurves': self.key_share_curves,
            }
        else:
            request_payload['tlsClientIdentifier'] = self.client_identifier
            request_payload['withRandomTLSExtensionOrder'] = self.random_tls_extension_order

        return request_payload, headers

    def build_response_obj(
        self,
        url: str,
        headers: Optional[Union[dict, CaseInsensitiveDict]],
        response_object: dict,
    ):
        if response_object['status'] == 0:
            raise ClientException(response_object['body'])
        # Set response cookies
        response_cookie_jar = extract_cookies_to_jar(
            request_url=url,
            request_headers=headers,
            cookie_jar=self.cookies,
            response_headers=response_object['headers'],
        )
        # build response class
        return playfulrequests.response.build_response(response_object, response_cookie_jar)

    def build_response(
        self,
        url,
        headers: Optional[Union[dict, CaseInsensitiveDict]],
        response_object: dict,
    ):  # sourcery skip: assign-if-exp
        history: list = []
        if not response_object['isHistory']:
            return self.build_response_obj(url, headers, response_object['response'])
        resps: list = response_object['history']
        for index, item in enumerate(resps):
            if index:  # > 0
                # get the location redirect url from the previous response
                item_url = resps[index - 1]['headers']['Location'][0]
            else:
                # use the original url
                item_url = url
            history.append(self.build_response_obj(item_url, headers, item))
        # assign history to last response
        resp = history[-1]
        resp.history = history[:-1]
        return resp

    def execute_request(
        self,
        method: str,
        url: str,
        headers: Optional[Union[dict, CaseInsensitiveDict]] = None,
        *args,
        **kwargs,
    ):
        '''
        execute a single request and return a response object
        '''
        # build request payload
        request_payload, headers = self.build_request(method, url, headers, *args, **kwargs)
        try:
            # send request
            resp = self.server.post(f'http://127.0.0.1:{PORT}/request', body=dumps(request_payload))
            response_object = loads(resp.read())
        except Exception as e:
            raise ClientException('Request failed') from e
        # build response class
        return self.build_response(url, headers, response_object)



from functools import partial
from typing import Dict, List, Tuple

from async_class import AsyncObject
from playwright.async_api import async_playwright

from playfulrequests.extensions import LoadFirefoxAddon
from playfulrequests.playwright_mock import Faker, ProxyManager, context
from playfulrequests.cffi import GetOpenPort


class PlaywrightMockBase(AsyncObject):
    async def __ainit__(
        self,
        headless: bool = False,
        scroll_into_view: bool = True,
        extensions=None,
    ):
        # setting values
        self.scroll_into_view = scroll_into_view
        self.extensions = extensions
        # starting Playwright
        self.playwright = await async_playwright().start()
        # launching chromium
        self.main_browser = await self.launch_browser(headless=headless)

    args: Tuple[str]

    async def stop(self):
        await self.main_browser.close()
        await self.playwright.stop()

    async def new_context(self, browser_name: str, user_agent: str, proxy=None, **launch_args):
        # calling proxyManager and faker
        _proxy = await ProxyManager(self, proxy)
        _faker = await Faker(self, _proxy, browser_name, user_agent)
        # create context with human emulation
        _browser = await context.new_context(
            self, _proxy, _faker, bypass_csp=True, user_agent=user_agent, **launch_args
        )
        _browser.proxy = _proxy
        _browser.faker = _faker

        return _browser


class ChromeBrowser(PlaywrightMockBase):
    args: Tuple[str] = (
        '--disable-blink-features=AutomationControlled',
        '--disable-web-security',
        '--disable-site-isolation-trials',
        '--disable-features=CrossSiteDocumentBlockingIfIsolating,'
        'CrossSiteDocumentBlockingAlways,'
        'IsolateOrigins,'
        'site-per-process,'
        'SharedArrayBuffer',
    )  # type: ignore

    async def launch_browser(self, headless: bool = False):
        args: List[str] = list(self.args)
        if headless:
            args.append('--headless=new')
        if self.extensions:
            paths = [ext.path for ext in self.extensions]
            args.extend(f'--load-extension={ext}' for ext in paths)
            args.append(f'--disable-extensions-except={",".join(paths)}')
        return await self.playwright.chromium.launch(headless=headless, args=args)


class FirefoxBrowser(PlaywrightMockBase):
    firefox_user_prefs: Dict[str, bool] = {
        'media.peerconnection.enabled': False,
        'media.navigator.enabled': False,
        'privacy.resistFingerprinting': False,
        'devtools.debugger.remote-enabled': True,
        'devtools.debugger.prompt-connection': False,
        'extensions.manifestV3.enabled': True,
    }

    async def launch_browser(self, headless: bool = False):
        run_cmd = partial(
            self.playwright.firefox.launch,
            headless=headless,
            firefox_user_prefs=self.firefox_user_prefs,
        )
        if not self.extensions:
            return await run_cmd()
        # block mv3
        for ext in self.extensions:
            assert not ext.is_mv3, f'MKV extensions are not supported: {ext.path}'
        rdp_port: int = GetOpenPort()
        pr = await run_cmd(
            args=('-start-debugger-server', str(rdp_port)),
        )
        for ext in self.extensions:
            ff_addon = LoadFirefoxAddon(rdp_port, ext.path)
            await ff_addon.load()
        return pr



import contextlib
import weakref
from functools import partial
from typing import List, MutableMapping, Optional, Set, Union
from urllib.parse import urljoin, urlparse, urlunparse

import selectolax.lexbor
from parse import Result, findall
from parse import search as parse_search

import playfulrequests
from playfulrequests.exceptions import NotRenderedException, SelectorNotFoundException

"""
Parser structure inspired by https://github.com/psf/requests-html/blob/master/requests_html.py
Copyright 2018 Kenneth Reitz
"""


DEFAULT_URL = 'https://example.org/'
DEFAULT_NEXT_SYMBOL = ['next', 'more', 'older']


# Typing
_Find = Union[List['Element'], 'Element']
_XPath = Union[List[str], List['Element'], str, 'Element']
_Result = Union[List['Result'], 'Result']
_HTML = Union[str, bytes]
_Containing = Union[str, List[str]]
_Links = Set[str]
_Next = Union['HTML', List[str]]
_NextSymbol = List[str]


class BaseParser:
    """
    A basic HTML/Element Parser, for Humans.

    Args:
        element: The element from which to base the parsing upon.
        html: HTML from which to base the parsing upon (optional).
        url: The URL from which the HTML originated, used for ``absolute_links``.
    """

    def __init__(
        self,
        *,
        element,
        url: str,
        br_session: Optional[weakref.CallableProxyType] = None,
    ) -> None:
        self.element = element
        self.url = url
        self.skip_anchors = True
        self.br_session = br_session

    # BrowserSession methods that accept a `selector` parameter
    pass_to_session = {
        'awaitSelector',
        'awaitEnabled',
        'isVisible',
        'isEnabled',
        'dragTo',
        'type',
        'click',
        'hover',
        'screenshot',
    }

    def __getattr__(self, name):
        if name not in self.pass_to_session:
            return
        # pass through to session if it exists
        if self.br_session is None:
            raise NotRenderedException(f'Method {name} only allowed in BrowserSession')
        return partial(getattr(self.br_session, name), self.css_path)

    @property
    def css_path(self) -> str:
        '''
        Returns the CSS selector of the element
        '''
        node = self.element
        elems: List[str] = []
        count: int
        while True:
            count = 0
            target_id = node.mem_id
            if node.tag == 'html' or not node.parent:
                break
            for elem in node.parent.iter():
                if elem.tag == node.tag:
                    count += 1
                    if elem.mem_id == target_id:
                        elems.append(f'{node.tag}:nth-of-type({count})')
                        node = node.parent
                        break
        return ' > '.join(elems[::-1])

    @property
    def raw_html(self) -> bytes:
        """
        Bytes representation of the HTML content.
        (`learn more <http://www.diveintopython3.net/strings.html>`_).
        """
        return self.element.raw_html

    @property
    def html(self) -> str:
        """
        Unicode representation of the HTML content
        (`learn more <http://www.diveintopython3.net/strings.html>`_).
        """
        return self.element.html

    @property
    def text(self) -> str:
        """
        The text content of the
        :class:`Element <Element>` or :class:`HTML <HTML>`.
        """
        return self.get_text()

    def get_text(self, children=True, separator='\n', strip=False) -> str:
        '''
        Get the text of the element, including its children.
        Args:
            children: Whether or not to include the children.
            separator: The separator to use between the text of the children.
            strip: Whether or not to strip the text.
        '''
        return self.element.text(separator=separator, strip=strip, deep=children)

    @property
    def full_text(self) -> str:
        """
        The full text content (including links) of the
        :class:`Element <Element>` or :class:`HTML <HTML>`.
        """
        return self.element.text_content

    kwarg_map = {
        'class_': 'class',
        'for_': 'for',
        'async_': 'async',
        'accept_charset': 'accept-charset',
        'http_equiv': 'http-equiv',
    }

    def find_all(
        self,
        selector: str = "*",
        other_kwargs: Optional[MutableMapping] = None,
        *,
        containing: _Containing = None,
        first: bool = False,
        **kwargs,
    ) -> _Find:
        """
        Given a CSS Selector, returns a list of
        :class:`Element <Element>` objects or a single one.

        Args:
            selector: CSS Selector to use.
            clean: Whether or not to sanitize the found HTML of ``<script>`` and ``<style>`` tags.
            containing: If specified, only return elements that contain the provided text.
            first: Whether or not to return just the first result.

        Returns:
            A list of :class:`Element <Element>` objects or a single one.

        Example CSS Selectors:
        - ``a``
        - ``a.someClass``
        - ``a#someID``
        - ``a[target=_blank]``
        See W3School's `CSS Selectors Reference
        <https://www.w3schools.com/cssref/css_selectors.asp>`_
        for more details.
        If ``first`` is ``True``, only returns the first
        :class:`Element <Element>` found.
        """

        # convert a single containing into a list
        if isinstance(containing, str):
            containing = [containing]

        # merge kwargs into a selector string
        if kwargs or other_kwargs:
            # map kwargs that resemble built-in attributes to their css equivalents
            for k, v in self.kwarg_map.items():
                if k in kwargs:
                    kwargs[v] = kwargs.pop(k)
            if other_kwargs:  # add passed dict to kwargs
                kwargs.update(other_kwargs)
            selector = selector.rstrip() + ''.join(f'[{k}="{v}"]' for k, v in kwargs.items())

        # find elements
        if first:
            found_css = (self.element.css_first(selector),)
            # if no element was found, raise an exception
            if not found_css[0]:
                raise SelectorNotFoundException(
                    f"No elements were found with selector '{selector}'."
                )
        else:
            found_css = self.element.css(selector)

        elements = [
            Element(element=found, url=self.url, br_session=self.br_session) for found in found_css
        ]

        if containing:
            elements_copy = elements.copy()
            elements = [
                element
                for element in elements_copy
                if any(c.lower() in element.full_text.lower() for c in containing)
            ]
            elements.reverse()

        return _get_first_or_list(elements, first)

    def find(
        self,
        selector: str = "*",
        other_kwargs: Optional[MutableMapping] = None,
        *,
        exception_handler: Optional[callable] = None,
        containing: _Containing = None,
        **kwargs,
    ) -> _Find:
        # Wrapper around find_all with first=True.
        try:
            return self.find_all(
                selector=selector,
                other_kwargs=other_kwargs,
                containing=containing,
                first=True,
                **kwargs,
            )
        except SelectorNotFoundException:
            if exception_handler:
                return exception_handler()

    def search(self, template: str) -> Result:
        """
        Search the Element for the given Parse template.

        Args:
            template (str): The Parse template to use.

        Returns:
            Result: The result of the search.
        """
        return parse_search(template, self.html)

    def search_all(self, template: str) -> _Result:
        """
        Search the Element (multiple times) for the given parse template.

        Args:
            template (str): The Parse template to use.

        Returns:
            _Result: The result of the search.
        """
        return list(findall(template, self.html))

    @property
    def links(self) -> _Links:
        """All found links on page, in as-is form."""

        def gen():
            for link in self.find_all('a'):
                with contextlib.suppress(KeyError):
                    href = link.attrs['href'].strip()
                    if (
                        href
                        and not (href.startswith('#') and self.skip_anchors)
                        and not href.startswith(('javascript:', 'mailto:'))
                    ):
                        yield href

        return set(gen())

    def _make_absolute(self, link):
        """Makes a given link absolute."""

        # Parse the link with stdlib.
        parsed = urlparse(link)._asdict()

        # If link is relative, then join it with base_url.
        if not parsed['netloc']:
            return urljoin(self.base_url, link)

        # Link is absolute; if it lacks a scheme, add one from base_url.
        if not parsed['scheme']:
            parsed['scheme'] = urlparse(self.base_url).scheme

            # Reconstruct the URL to incorporate the new scheme.
            parsed = iter(parsed.values())
            return urlunparse(parsed)

        # Link is absolute and complete with scheme; nothing to be done here.
        return link

    @property
    def absolute_links(self) -> _Links:
        """
        All found links on page, in absolute form
        (`learn more <https://www.navegabem.com/absolute-or-relative-links.html>`_).
        """

        def gen():
            for link in self.links:
                yield self._make_absolute(link)

        return set(gen())

    @property
    def base_url(self) -> str:
        """
        The base URL for the page. Supports the ``<base>`` tag
        (`learn more <https://www.w3schools.com/tags/tag_base.asp>`_)."""

        if base := self.find_all('base', first=True):
            if result := base.attrs.get('href', '').strip():
                return result

        # Parse the url to separate out the path
        parsed = urlparse(self.url)._asdict()

        # Remove any part of the path after the last '/'
        parsed['path'] = '/'.join(parsed['path'].split('/')[:-1]) + '/'

        # Reconstruct the url with the modified path
        parsed = iter(parsed.values())
        url = urlunparse(parsed)

        return url


class Element(BaseParser):
    """
    An element of HTML.

    Args:
        element: The element from which to base the parsing upon.
        url: The URL from which the HTML originated, used for ``absolute_links``.
    """

    __slots__ = [
        'element',
        'url',
        'skip_anchors',
        '_attrs',
        'session',
    ]

    def __init__(
        self,
        *,
        element,
        url: str,
        br_session: Optional[weakref.CallableProxyType] = None,
    ) -> None:
        super(Element, self).__init__(element=element, url=url, br_session=br_session)
        self.element = element
        self.tag = element.tag
        self._attrs = None

    def __repr__(self) -> str:
        attrs = [f'{attr}={repr(self.attrs[attr])}' for attr in self.attrs]
        return f"<Element {repr(self.element.tag)} {' '.join(attrs)}>"

    @property
    def attrs(self) -> MutableMapping:
        """
        Returns a dictionary of the attributes of the :class:`Element <Element>`
        (`learn more <https://www.w3schools.com/tags/ref_attributes.asp>`_).
        """
        if self._attrs is None:
            self._attrs = self.element.attributes

            # Split class and rel up, as there are usually many of them:
            for attr in ['class', 'rel']:
                if attr in self._attrs:
                    self._attrs[attr] = tuple(self._attrs[attr].split())

        return self._attrs

    def __getattr__(self, name):
        if name in self.attrs:
            return self.attrs[name]
        if name in self.kwarg_map:
            return self.attrs[self.kwarg_map[name]]
        return super().__getattr__(name)


class HTML(BaseParser):
    """An HTML document, ready for parsing.

    Args:
        url (str): The URL from which the HTML originated, used for ``absolute_links``.
        html (Optional[_HTML]): HTML from which to base the parsing upon.

    Attributes:
        session (Union[playfulrequest.session.TLSSession, playfulrequest.browser.BrowserSession]): The session used for the HTML request.
    """

    def __init__(
        self,
        *,
        session: Optional[
            Union['playfulrequests.session.TLSSession', 'playfulrequests.browser.BrowserSession']
        ] = None,
        url: str = DEFAULT_URL,
        html: _HTML,
    ) -> None:
        # Convert incoming unicode HTML into bytes.
        element = selectolax.lexbor.LexborHTMLParser(html)
        super(HTML, self).__init__(
            element=element,
            url=url,
            br_session=weakref.proxy(session)
            if hasattr(playfulrequests, 'browser')  # if the browser module is imported
               and isinstance(
                session, playfulrequests.browser.BrowserSession
            )  # and session is a BrowserSession
            else None,
        )
        self.session = session or playfulrequests.firefox.Session(temp=True)
        self.page = None
        self.next_symbol = DEFAULT_NEXT_SYMBOL

    def __repr__(self) -> str:
        return f"<HTML url={self.url!r}>"

    def next(self, fetch: bool = False, next_symbol: _NextSymbol = None) -> _Next:
        """
        Attempts to find the next page, if there is one. If ``fetch``
        is ``True`` (default), returns :class:`HTML <HTML>` object of
        next page. If ``fetch`` is ``False``, simply returns the next URL.
        """
        if next_symbol is None:
            next_symbol = DEFAULT_NEXT_SYMBOL

        def get_next():
            candidates = self.find_all('a', containing=next_symbol)

            for candidate in candidates:
                if candidate.attrs.get('href'):
                    # Support 'next' rel (e.g. reddit).
                    if 'next' in candidate.attrs.get('rel', []):
                        return candidate.attrs['href']

                    # Support 'next' in classnames.
                    for _class in candidate.attrs.get('class', []):
                        if 'next' in _class:
                            return candidate.attrs['href']

                    if 'page' in candidate.attrs['href']:
                        return candidate.attrs['href']

            try:
                # Resort to the last candidate.
                return candidates[-1].attrs['href']
            except IndexError:
                return None

        __next = get_next()
        if __next:
            url = self._make_absolute(__next)
        else:
            return None

        return self.session.get(url) if fetch else url

    def __iter__(self):
        next_ = self
        while True:
            yield next_
            try:
                next_ = next_.next(fetch=True, next_symbol=self.next_symbol).html
            except AttributeError:
                break

    def __next__(self):
        return self.next(fetch=True, next_symbol=self.next_symbol).html

    def __aiter__(self):
        return self

    async def __anext__(self):
        while True:
            url = self.next(fetch=False, next_symbol=self.next_symbol)
            if not url:
                break
            response = await self.session.get(url)
            return response.html

    def add_next_symbol(self, next_symbol):
        self.next_symbol.append(next_symbol)


def _get_first_or_list(l, first=False):
    if not first:
        return l
    try:
        return l[0]
    except IndexError:
        return None



import typing

from . import element_handle, frame, frame_locator, js_handle


def attach_dyn_propr(instance, prop_name, propr):
    """Attach property proper to instance with name prop_name.

    Reference:
      * https://stackoverflow.com/a/1355444/509706
      * https://stackoverflow.com/questions/48448074
    """
    class_name = f"{instance.__class__.__name__}Child"
    child_class = type(class_name, (instance.__class__,), {prop_name: propr})

    instance.__class__ = child_class


def mock_locator(locator) -> None:
    async def click_mocker(
        button="left",
        click_count=1,
        delay=20,
        force=False,
        modifiers=[],
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await Locator.click(
            locator,
            button=button,
            click_count=click_count,
            delay=delay,
            force=force,
            modifiers=modifiers,
            no_wait_after=no_wait_after,
            position=position,
            timeout=timeout,
            trial=trial,
        )

    locator.click = click_mocker

    async def dblclick_mocker(
        button="left",
        delay=20,
        force=False,
        modifiers=[],
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await Locator.dblclick(
            locator,
            button=button,
            delay=delay,
            force=force,
            modifiers=modifiers,
            no_wait_after=no_wait_after,
            position=position,
            timeout=timeout,
            trial=trial,
        )

    locator.dblclick = dblclick_mocker

    async def check_mocker(
        force=False,
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await Locator.check(
            locator,
            force=force,
            no_wait_after=no_wait_after,
            position=position,
            timeout=timeout,
            trial=trial,
        )

    locator.check = check_mocker

    async def uncheck_mocker(
        force=False,
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await Locator.uncheck(
            locator,
            force=force,
            no_wait_after=no_wait_after,
            position=position,
            timeout=timeout,
            trial=trial,
        )

    locator.uncheck = uncheck_mocker

    async def set_checked_mocker(
        checked=False,
        force=False,
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await Locator.set_checked(
            locator,
            checked=checked,
            force=force,
            no_wait_after=no_wait_after,
            position=position,
            timeout=timeout,
            trial=trial,
        )

    locator.set_checked = set_checked_mocker

    async def hover_mocker(
        force=False, modifiers=[], position={}, timeout: typing.Optional[float] = None, trial=False
    ):
        await Locator.hover(
            locator,
            force=force,
            modifiers=modifiers,
            position=position,
            timeout=timeout,
            trial=trial,
        )

    locator.hover = hover_mocker

    async def type_mocker(
        text, delay=200, no_wait_after=False, timeout: typing.Optional[float] = None
    ):
        await Locator.type(locator, text, delay=delay, no_wait_after=no_wait_after, timeout=timeout)

    locator.type = type_mocker

    # JsHandle
    async def mock_evaluate_handle(expression, arg=None) -> "JSHandle":
        _js_handle = await locator._evaluate_handle(expression, arg=arg)
        await js_handle.mock_js_handle(_js_handle, locator.page)
        return _js_handle

    locator._evaluate_handle = locator.evaluate_handle
    locator.evaluate_handle = mock_evaluate_handle

    # FrameLocator
    async def mock_frame_locator_func(selector) -> "JSHandle":
        _frame_locator = await locator._frame_locator(selector)
        await frame_locator.mock_frame_locator(_frame_locator)
        return _frame_locator

    locator._frame_locator = locator.frame_locator
    locator.frame_locator = mock_frame_locator_func

    # ElementHandle
    def element_handle_mocker(timeout: typing.Optional[float] = None):
        element = locator._element_handle(timeout=timeout)
        element_handle.mock_element_handle(element, locator.page)
        return element

    locator._element_handle = locator.element_handle
    locator.element_handle = element_handle_mocker

    # Locator
    def nth_mocker(index):
        _locator = locator._nth(index)
        mock_locator(_locator)
        return _locator

    locator._nth = locator.nth
    locator.nth = nth_mocker

    @property
    def first_mocker(self):
        _locator = locator._first
        mock_locator(_locator)
        return _locator

    @property
    def last_mocker(self):
        _locator = locator._last
        mock_locator(_locator)
        return _locator

    locator._first = locator.first
    locator._last = locator.last

    attach_dyn_propr(locator, "first", first_mocker)
    attach_dyn_propr(locator, "last", last_mocker)


class Locator:
    async def click(
        locator,
        button="left",
        click_count=1,
        delay=20,
        force=False,
        modifiers=[],
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        if not force:
            await locator.wait_for(state="visible", timeout=timeout)

        if not trial:
            if locator.page.scroll_into_view:
                await locator.scroll_into_view_if_needed(timeout=timeout)

            boundings = await locator.bounding_box()
            x, y, width, height = boundings.values()

            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            for modifier in modifiers:
                await locator.page.keyboard.down(modifier)

            await locator.page.mouse.click(x, y, button, click_count, delay)

            for modifier in modifiers:
                await locator.page.keyboard.up(modifier)

    async def dblclick(
        locator,
        button="left",
        delay=20,
        force=False,
        modifiers=[],
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        if not force:
            await locator.wait_for(state="visible", timeout=timeout)

        if not trial:
            if locator.page.scroll_into_view:
                await locator.scroll_into_view_if_needed(timeout=timeout)

            boundings = await locator.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            for modifier in modifiers:
                await locator.page.keyboard.down(modifier)

            await locator.page.mouse.dblclick(x, y, button, delay)

            for modifier in modifiers:
                await locator.page.keyboard.up(modifier)

    async def check(
        locator,
        force=False,
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        if not force:
            await locator.wait_for(state="visible", timeout=timeout)

        if await locator.is_checked():
            return

        if not trial:
            if locator.page.scroll_into_view:
                await locator.scroll_into_view_if_needed(timeout=timeout)

            boundings = await locator.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            await locator.page.mouse.click(x, y, button="left", click_count=1, delay=20)

            assert await locator.is_checked()

    async def uncheck(
        locator,
        force=False,
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        if not force:
            await locator.wait_for(state="visible", timeout=timeout)

        if not await locator.is_checked():
            return

        if not trial:
            if locator.page.scroll_into_view:
                await locator.scroll_into_view_if_needed(timeout=timeout)

            boundings = await locator.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            await locator.page.mouse.click(x, y, button="left", click_count=1, delay=20)

            assert not await locator.is_checked()

    async def set_checked(
        locator,
        checked=False,
        force=False,
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        if not force:
            await locator.wait_for(state="visible", timeout=timeout)

        if await locator.is_checked() == checked:
            return

        if not trial:
            if locator.page.scroll_into_view:
                await locator.scroll_into_view_if_needed(timeout=timeout)

            boundings = await locator.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            await locator.page.mouse.click(x, y, button="left", click_count=1, delay=20)

            assert await locator.is_checked()

    async def hover(
        locator,
        force=False,
        modifiers=[],
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        if not force:
            await locator.wait_for(state="visible", timeout=timeout)

        if not trial:
            if locator.page.scroll_into_view:
                await locator.scroll_into_view_if_needed(timeout=timeout)

            boundings = await locator.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            for modifier in modifiers:
                await locator.page.keyboard.down(modifier)

            await locator.page.mouse.move(x, y)

            for modifier in modifiers:
                await locator.page.keyboard.up(modifier)

    async def type(
        locator, text, delay=200, no_wait_after=False, timeout: typing.Optional[float] = None
    ) -> None:
        await locator.wait_for(state="visible", timeout=timeout)

        if locator.page.scroll_into_view:
            await locator.scroll_into_view_if_needed(timeout=timeout)

        boundings = await locator.bounding_box()
        x, y, width, height = boundings.values()

        x, y = x + width // 2, y + height // 2

        await locator.page.mouse.click(x, y, "left", 1, delay)

        await locator.page.keyboard.type(text, delay=delay)



import random
import typing

import playwright_stealth

from . import element_handle, frame, frame_locator, js_handle, locator, mouse


async def new_page(inst, context, proxy, faker, **launch_arguments) -> "PlaywrightPage":
    # Create new Page
    page = await context._new_page(**launch_arguments)
    # Stealthen the page with custom Stealth Config
    config = playwright_stealth.StealthConfig()
    (
        config.navigator_languages,
        config.permissions,
        config.navigator_platform,
        config.navigator_vendor,
        config.outerdimensions,
    ) = (False, False, False, False, False)
    # Setting Important JS Variables to inst Variables
    config.vendor, config.renderer, config.nav_user_agent, config.nav_platform = (
        faker.vendor,
        faker.renderer,
        faker.useragent,
        "Win32",
    )
    # Setting the Language
    config.languages = ("en-US", "en", faker.locale, faker.language_code)
    await playwright_stealth.stealth_async(page, config)

    # Mocking Page
    await mock_page(page, inst)
    return page


async def mock_keyboard(page) -> None:
    # KeyboardMocking
    async def type_mocker(text, delay=50) -> None:
        offset = delay // 2
        for char in text:
            # randomize the delay by 50% innacuracy
            await page.keyboard._type(
                char, delay=max(random.randint(delay - offset, delay + offset), 0)
            )
        await page.wait_for_timeout(random.randint(4, 8) * 100)

    page.keyboard._type = page.keyboard.type
    page.keyboard.type = type_mocker


async def mock_page_functions(page):
    # Frame
    def mock_frame_func(name=None, url=None) -> "Frame":
        _frame = page._frame(name=name, url=url)
        frame.mock_frame(_frame)
        return _frame

    page._frame = page.frame
    page.frame = mock_frame_func

    # ElementHandle
    async def mock_query_selector(selector, strict=False) -> typing.Optional["ElementHandle"]:
        element = await page._query_selector(selector, strict=strict)
        if element:
            element_handle.mock_element_handle(element, page)
        return element

    page._query_selector = page.query_selector
    page.query_selector = mock_query_selector

    async def mock_query_selector_all(selector) -> typing.List["ElementHandle"]:
        elements = await page._query_selector_all(selector)
        for element in elements:
            element_handle.mock_element_handle(element, page)
        return elements

    page._query_selector_all = page.query_selector_all
    page.query_selector_all = mock_query_selector_all

    async def mock_wait_for_selector(
        selector, state=[], strict=False, timeout: typing.Optional[float] = None
    ) -> typing.Optional["ElementHandle"]:
        element = await page._wait_for_selector(
            selector, state=state, strict=strict, timeout=timeout
        )
        if element:
            element_handle.mock_element_handle(element, page)
        return element

    page._wait_for_selector = page.wait_for_selector
    page.wait_for_selector = mock_wait_for_selector

    async def mock_add_script_tag(content="", path="", type="", url="") -> "ElementHandle":
        element = await page._add_script_tag(content=content, path=path, type=type, url=url)
        element_handle.mock_element_handle(element, page)
        return element

    page._add_script_tag = page.add_script_tag
    page.add_script_tag = mock_add_script_tag

    async def mock_add_style_tag(content="", path="", type="", url="") -> "ElementHandle":
        element = await page._add_script_tag(content=content, path=path, type=type, url=url)
        element_handle.mock_element_handle(element, page)
        return element

    page._add_script_tag = page.add_script_tag
    page.add_script_tag = mock_add_style_tag

    # Locator
    def mock_locator_func(selector, has=None, has_text="") -> "Locator":
        _locator = page._locator(selector, has=has, has_text=has_text)
        locator.mock_locator(_locator)
        return _locator

    page._locator = page.locator
    page.locator = mock_locator_func

    # JsHandle
    async def mock_evaluate_handle(expression, arg=None) -> "JSHandle":
        _js_handle = await page._evaluate_handle(expression, arg=arg)
        js_handle.mock_js_handle(_js_handle, page)
        return _js_handle

    page._evaluate_handle = page.evaluate_handle
    page.evaluate_handle = mock_evaluate_handle

    async def mock_wait_for_function(
        expression, arg=None, polling="raf", timeout: typing.Optional[float] = None
    ) -> "JSHandle":
        _js_handle = await page._wait_for_function(
            expression, arg=arg, polling=polling, timeout=timeout
        )
        js_handle.mock_js_handle(_js_handle, page)
        return _js_handle

    page._wait_for_function = page.wait_for_function
    page.wait_for_function = mock_wait_for_function

    # FrameLocator
    def mock_frame_locator_func(selector) -> "JSHandle":
        _frame_locator = page._frame_locator(selector)
        frame_locator.mock_frame_locator(_frame_locator)
        return _frame_locator

    page._frame_locator = page.frame_locator
    page.frame_locator = mock_frame_locator_func


async def mock_page_objects(page) -> None:
    async def click_mocker(
        selector,
        button="left",
        click_count=1,
        strict=False,
        delay=20,
        force=False,
        modifiers=[],
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await Page.click(
            page,
            selector,
            button=button,
            click_count=click_count,
            strict=strict,
            delay=delay,
            force=force,
            modifiers=modifiers,
            no_wait_after=no_wait_after,
            position=position,
            timeout=timeout,
            trial=trial,
        )

    page.click = click_mocker

    async def dblclick_mocker(
        selector,
        button="left",
        strict=False,
        delay=20,
        force=False,
        modifiers=[],
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await Page.dblclick(
            page,
            selector,
            button=button,
            strict=strict,
            delay=delay,
            force=force,
            modifiers=modifiers,
            no_wait_after=no_wait_after,
            position=position,
            timeout=timeout,
            trial=trial,
        )

    page.dblclick = dblclick_mocker

    async def check_mocker(
        selector,
        force=False,
        no_wait_after=False,
        position={},
        strict=False,
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await Page.check(
            page,
            selector,
            force=force,
            no_wait_after=no_wait_after,
            position=position,
            strict=strict,
            timeout=timeout,
            trial=trial,
        )

    page.check = check_mocker

    async def uncheck_mocker(
        selector,
        force=False,
        no_wait_after=False,
        position={},
        strict=False,
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await page.uncheck(
            page,
            selector,
            force=force,
            no_wait_after=no_wait_after,
            position=position,
            strict=strict,
            timeout=timeout,
            trial=trial,
        )

    page.uncheck = uncheck_mocker

    async def set_checked_mocker(
        selector,
        checked=False,
        force=False,
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await Page.set_checked(
            page,
            selector,
            checked=checked,
            force=force,
            no_wait_after=no_wait_after,
            position=position,
            timeout=timeout,
            trial=trial,
        )

    page.set_checked = set_checked_mocker

    async def hover_mocker(
        selector,
        force=False,
        modifiers=[],
        position={},
        strict=False,
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await Page.hover(
            page,
            selector,
            force=force,
            modifiers=modifiers,
            position=position,
            strict=strict,
            timeout=timeout,
            trial=trial,
        )

    page.hover = hover_mocker

    async def type_mocker(
        selector,
        text,
        delay=50,
        no_wait_after=False,
        strict=False,
        timeout: typing.Optional[float] = None,
    ):
        await Page.type(
            page,
            selector,
            text,
            delay=delay,
            no_wait_after=no_wait_after,
            strict=strict,
            timeout=timeout,
        )

    page.type = type_mocker


class Page:
    async def click(
        page,
        selector,
        button="left",
        click_count=1,
        strict=False,
        delay=20,
        force=False,
        modifiers=[],
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        element = await page.wait_for_selector(
            selector, state="visible" if not force else "hidden", strict=strict, timeout=timeout
        )

        if not force:
            await element.wait_for_element_state("editable", timeout=timeout)

        if not trial:
            # await element.scroll_into_view_if_needed(timeout=timeout)

            boundings = await element.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            for modifier in modifiers:
                await page.keyboard.down(modifier)

            await page.mouse.click(x, y, button, click_count, delay)

            for modifier in modifiers:
                await page.keyboard.up(modifier)

    async def dblclick(
        page,
        selector,
        button="left",
        strict=False,
        delay=20,
        force=False,
        modifiers=[],
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        element = await page.wait_for_selector(
            selector, state="visible" if not force else "hidden", strict=strict, timeout=timeout
        )

        if not force:
            await element.wait_for_element_state("editable", timeout=timeout)

        if not trial:
            # await element.scroll_into_view_if_needed(timeout=timeout)

            boundings = await element.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            for modifier in modifiers:
                await page.keyboard.down(modifier)

            await page.mouse.dblclick(x, y, button, delay)

            for modifier in modifiers:
                await page.keyboard.up(modifier)

    async def check(
        page,
        selector,
        force=False,
        no_wait_after=False,
        position={},
        strict=False,
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        element = await page.wait_for_selector(
            selector, state="visible" if not force else "hidden", strict=strict, timeout=timeout
        )

        if not force:
            await element.wait_for_element_state("editable", timeout=timeout)

        if await element.is_checked():
            return

        if not trial:
            # await element.scroll_into_view_if_needed(timeout=timeout)

            boundings = await element.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            await page.mouse.click(x, y, button="left", click_count=1, delay=20)

            assert await element.is_checked()

    async def uncheck(
        page,
        selector,
        force=False,
        no_wait_after=False,
        position={},
        strict=False,
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        element = await page.wait_for_selector(
            selector, state="visible" if not force else "hidden", strict=strict, timeout=timeout
        )

        if not force:
            await element.wait_for_element_state("editable", timeout=timeout)

        if not await element.is_checked():
            return

        if not trial:
            # await element.scroll_into_view_if_needed(timeout=timeout)

            boundings = await element.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            await page.mouse.click(x, y, button="left", click_count=1, delay=20)

            assert not await element.is_checked()

    async def set_checked(
        page,
        selector,
        checked=False,
        force=False,
        no_wait_after=False,
        position={},
        strict=False,
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        element = await page.wait_for_selector(
            selector, state="visible" if not force else "hidden", strict=strict, timeout=timeout
        )

        if not force:
            await element.wait_for_element_state("editable", timeout=timeout)

        if await element.is_checked() == checked:
            return

        if not trial:
            # await element.scroll_into_view_if_needed(timeout=timeout)

            boundings = await element.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            await page.mouse.click(x, y, button="left", click_count=1, delay=20)

            assert await element.is_checked()

    async def hover(
        page,
        selector,
        force=False,
        modifiers=[],
        position={},
        strict=False,
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        element = await page.wait_for_selector(
            selector, state="visible" if not force else "hidden", strict=strict, timeout=timeout
        )

        if not force:
            await element.wait_for_element_state("editable", timeout=timeout)

        if not trial:
            # await element.scroll_into_view_if_needed(timeout=timeout)

            boundings = await element.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            for modifier in modifiers:
                await page.keyboard.down(modifier)

            await page.mouse.move(x, y)

            for modifier in modifiers:
                await page.keyboard.up(modifier)

    async def type(
        page,
        selector,
        text,
        delay=50,
        no_wait_after=False,
        strict=False,
        timeout: typing.Optional[float] = None,
    ) -> None:
        element = await page.wait_for_selector(
            selector, state="visible", strict=strict, timeout=timeout
        )

        await element.wait_for_element_state("editable", timeout=timeout)

        # await element.scroll_into_view_if_needed(timeout=timeout)

        boundings = await element.bounding_box()
        x, y, width, height = boundings.values()

        x, y = x + width // 2, y + height // 2

        await page.mouse.click(x, y, "left", 1, delay)

        await page.keyboard.type(text, delay=delay)


async def better_google_score(page):
    await page.goto("https://google.com/")
    await page.click('[id="L2AGLb"]')
    await page.type('[autocomplete="off"]', ".")
    await page.keyboard.press("Enter")
    await page.wait_for_timeout(1000)


async def mock_page(page, inst) -> None:
    # await better_google_score(page)

    page.scroll_into_view = inst.scroll_into_view

    mouse.mock_mouse(page)
    await mock_keyboard(page)

    frame.mock_frame(page.main_frame)
    await mock_page_objects(page)

    await mock_page_functions(page)



from . import locator


def attach_dyn_propr(instance, prop_name, propr):
    """Attach property proper to instance with name prop_name.

    Reference:
      * https://stackoverflow.com/a/1355444/509706
      * https://stackoverflow.com/questions/48448074
    """
    class_name = instance.__class__.__name__ + "Child"
    child_class = type(class_name, (instance.__class__,), {prop_name: propr})

    instance.__class__ = child_class


def mock_frame_locator(frame_locator) -> None:
    # MouseMocking
    def locator_mocker(selector, has=None, has_text=""):
        _locator = frame_locator._locator(selector, has=has, has_text=has_text)
        locator.mock_locator(_locator)
        return _locator

    frame_locator._locator = frame_locator.locator
    frame_locator.locator = locator_mocker

    # FrameLocator
    def nth_mocker(index):
        _locator = frame_locator._nth(index)
        mock_frame_locator(_locator)
        return _locator

    frame_locator._nth = frame_locator.nth
    frame_locator.nth = nth_mocker

    @property
    def first_mocker(self):
        _locator = frame_locator._first
        mock_frame_locator(_locator)
        return _locator

    @property
    def last_mocker(self):
        _locator = frame_locator._last
        mock_frame_locator(_locator)
        return _locator

    frame_locator._first = frame_locator.first
    frame_locator._last = frame_locator.last

    attach_dyn_propr(frame_locator, "first", first_mocker)
    attach_dyn_propr(frame_locator, "last", last_mocker)



import os
from collections import OrderedDict
from collections.abc import Mapping
from dataclasses import dataclass
from io import BufferedReader
from typing import Any, Dict, Iterator, List, Mapping, MutableMapping, Optional, Tuple, Union

from urllib3.fields import RequestField
from urllib3.filepost import encode_multipart_formdata


class FileUtils:
    basestring = (str, bytes)

    @staticmethod
    def to_items_list(value: Union[dict, tuple, list]) -> List[Tuple[str, Any]]:
        if isinstance(value, Mapping):
            value = value.items()
        return list(value)

    @staticmethod
    def _guess_filename(obj) -> Optional[str]:
        '''Tries to guess the filename of the given object.'''
        name = getattr(obj, 'name', None)
        if name and isinstance(name, FileUtils.basestring) and name[0] != "<" and name[-1] != ">":
            return os.path.basename(name)  # type: ignore

    @staticmethod
    def get_fields(data: Union[dict, tuple]) -> Iterator[Tuple[str, bytes]]:
        fields = FileUtils.to_items_list(data)
        for field, val in fields:
            if isinstance(val, FileUtils.basestring) or not hasattr(val, "__iter__"):
                val = (val,)
            for v in val:
                if v is None:
                    continue
                if not isinstance(v, bytes):
                    v = str(v)
                yield (
                    field.decode("utf-8") if isinstance(field, bytes) else field,
                    v.encode("utf-8") if isinstance(v, str) else v,
                )

    @staticmethod
    def encode_files(
        files: Dict[str, Union[BufferedReader, tuple]], data: Optional[Union[dict, tuple]] = None
    ) -> Tuple[bytes, str]:
        """
        Build the body for a multipart/form-data request.
        Will encode files when passed as a dict or a list of tuples.

        (file_name, file_obj[, content_type[, custom_headers]])
        """
        fields = list(FileUtils.get_fields(data)) if data else []

        for k, v in FileUtils.to_items_list(files):
            # support for explicit filename
            if isinstance(v, (tuple, list)):
                assert len(v) in range(
                    2, 5
                ), "Tuple/list must be (file_name, file_obj[, content_type[, custom_headers]])"
                file = File(*v)
            else:
                file = File(FileUtils._guess_filename(v) or k, v)

            rf: RequestField = RequestField(
                name=k, data=file.fdata, filename=file.file_name, headers=file.custom_headers
            )
            rf.make_multipart(content_type=file.content_type)
            fields.append(rf)

        return encode_multipart_formdata(fields)


@dataclass
class File:
    file_name: str
    file_obj: Union[BufferedReader, str, bytes, bytearray]
    content_type: Optional[str] = None
    custom_headers: Optional[dict] = None

    def __post_init__(self) -> None:
        # get the file data
        if isinstance(self.file_obj, BufferedReader):
            self.fdata = self.file_obj.read()
        else:
            self.fdata = self.file_obj


class CaseInsensitiveDict(MutableMapping):
    '''
    Origin: requests library (https://github.com/psf/requests)
    A case-insensitive ``dict``-like object.
    '''

    def __init__(self, data=None, **kwargs):
        self._store = OrderedDict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def __setitem__(self, key, value):
        # Use the lowercased key for lookups, but store the actual
        # key alongside the value.
        self._store[key.lower()] = (key, value)

    def __getitem__(self, key):
        return self._store[key.lower()][1]

    def __delitem__(self, key):
        del self._store[key.lower()]

    def __iter__(self):
        return (casedkey for casedkey, mappedvalue in self._store.values())

    def __len__(self):
        return len(self._store)

    def lower_items(self):
        '''Like iteritems(), but with all lowercase keys.'''
        return ((lowerkey, keyval[1]) for (lowerkey, keyval) in self._store.items())

    def __eq__(self, other):
        if isinstance(other, Mapping):
            other = CaseInsensitiveDict(other)
        else:
            return NotImplemented
        # Compare insensitively
        return dict(self.lower_items()) == dict(other.lower_items())

    # Copy is required
    def copy(self):
        return CaseInsensitiveDict(self._store.values())

    def __repr__(self):
        return str(dict(self.items()))



from __future__ import annotations

import math
import random
from typing import Optional, Tuple, List, Any

import numpy as np
from playwright.async_api import Page


def mock_mouse(page: Page) -> None:
    mouse_mocker = MouseMock(page)

    # MouseMocking
    async def click_mocker(
        x: int | float,
        y: int | float,
        button: Optional[str] = "left",
        click_count: Optional[int] = 1,
        delay: Optional[int] = 20,
        humanly: Optional[bool] = True,
    ) -> None:
        await mouse_mocker.click(x, y, button, click_count, delay, humanly)

    page.mouse.click = click_mocker

    async def dblclick_mocker(
        x: int | float,
        y: int | float,
        button: Optional[str] = "left",
        delay: Optional[int] = 20,
        humanly: Optional[bool] = True,
    ) -> None:
        await mouse_mocker.dblclick(x, y, button, delay, humanly)

    page.mouse.dblclick = dblclick_mocker

    async def move_mocker(
        x: int | float, y: int | float, steps: Optional[int] = 1, humanly: Optional[bool] = True
    ) -> None:
        await mouse_mocker.move(x, y, steps, humanly)

    page.mouse.origin_move = page.mouse.move
    page.mouse.move = move_mocker


# From https://github.com/riflosnake/HumanCursor/blob/main/humancursor/utilities/human_curve_generator.py
class HumanizeMouseTrajectory:
    def __init__(
        self, from_point: Tuple[int | float, int | float], to_point: Tuple[int | float, int | float]
    ) -> None:
        self.from_point = from_point
        self.to_point = to_point
        self.points = self.generate_curve()

    def easeOutQuad(self, n: float) -> float:
        if not 0.0 <= n <= 1.0:
            raise ValueError('Argument must be between 0.0 and 1.0.')
        return -n * (n - 2)

    def generate_curve(self) -> List[Tuple[int, int]]:
        """Generates the curve based on arguments below, default values below are automatically modified to cause randomness"""
        left_boundary = min(self.from_point[0], self.to_point[0]) - 80
        right_boundary = max(self.from_point[0], self.to_point[0]) + 80
        down_boundary = min(self.from_point[1], self.to_point[1]) - 80
        up_boundary = max(self.from_point[1], self.to_point[1]) + 80

        tween = self.easeOutQuad

        internalKnots = self.generate_internal_knots(
            left_boundary, right_boundary, down_boundary, up_boundary, 2
        )
        points = self.generate_points(internalKnots)
        points = self.distort_points(points, 1, 1, 0.5)
        points = self.tween_points(points, 100)
        return points

    def generate_internal_knots(
        self,
        l_boundary: int | float,
        r_boundary: int | float,
        d_boundary: int | float,
        u_boundary: int | float,
        knots_count: int,
    ) -> List[Tuple[int, int]]:
        """Generates the internal knots of the curve randomly"""
        if not (
            self.check_if_numeric(l_boundary)
            and self.check_if_numeric(r_boundary)
            and self.check_if_numeric(d_boundary)
            and self.check_if_numeric(u_boundary)
        ):
            raise ValueError("Boundaries must be numeric values")
        if not isinstance(knots_count, int) or knots_count < 0:
            knots_count = 0
        if l_boundary > r_boundary:
            raise ValueError("left_boundary must be less than or equal to right_boundary")
        if d_boundary > u_boundary:
            raise ValueError("down_boundary must be less than or equal to upper_boundary")

        try:
            knotsX = np.random.choice(range(l_boundary, r_boundary) or l_boundary, size=knots_count)
            knotsY = np.random.choice(range(d_boundary, u_boundary) or d_boundary, size=knots_count)
        except TypeError:
            knotsX = np.random.choice(range(int(l_boundary), int(r_boundary)), size=knots_count)
            knotsY = np.random.choice(range(int(d_boundary), int(u_boundary)), size=knots_count)

        return list(zip(knotsX, knotsY))

    def generate_points(self, knots: List[Tuple[int, int]]) -> List[Tuple[int, int]]:
        """Generates the points from BezierCalculator"""
        if not self.check_if_list_of_points(knots):
            raise ValueError("knots must be valid list of points")

        midPtsCnt = max(
            abs(self.from_point[0] - self.to_point[0]),
            abs(self.from_point[1] - self.to_point[1]),
            2,
        )
        knots = [self.from_point] + knots + [self.to_point]
        return BezierCalculator.calculate_points_in_curve(int(midPtsCnt), knots)

    def distort_points(
        self,
        points: List[Tuple[int, int]],
        distortion_mean: int,
        distortion_st_dev: int,
        distortion_frequency: float,
    ) -> List[Tuple[int, int]]:
        """Distorts points by parameters of mean, standard deviation and frequency"""
        if not (
            self.check_if_numeric(distortion_mean)
            and self.check_if_numeric(distortion_st_dev)
            and self.check_if_numeric(distortion_frequency)
        ):
            raise ValueError("Distortions must be numeric")
        if not self.check_if_list_of_points(points):
            raise ValueError("points must be valid list of points")
        if not (0 <= distortion_frequency <= 1):
            raise ValueError("distortion_frequency must be in range [0,1]")

        distorted = []
        for i in range(1, len(points) - 1):
            x, y = points[i]
            delta = (
                np.random.normal(distortion_mean, distortion_st_dev)
                if random.random() < distortion_frequency
                else 0
            )
            distorted += ((x, y + delta),)
        distorted = [points[0]] + distorted + [points[-1]]
        return distorted

    def tween_points(
        self, points: List[Tuple[int, int]], target_points: int
    ) -> List[Tuple[int, int]]:
        """Modifies points by tween"""
        if not self.check_if_list_of_points(points):
            raise ValueError("List of points not valid")
        if not isinstance(target_points, int) or target_points < 2:
            raise ValueError("target_points must be an integer greater or equal to 2")

        res = []
        for i in range(target_points):
            index = int(self.easeOutQuad(float(i) / (target_points - 1)) * (len(points) - 1))
            res += (points[index],)
        return res

    @staticmethod
    def check_if_numeric(val: Any) -> bool:
        """Checks if value is proper numeric value"""
        return isinstance(val, (float, int, np.int32, np.int64, np.float32, np.float64))

    def check_if_list_of_points(self, list_of_points: List[Tuple[int, int]]) -> bool:
        """Checks if list of points is valid"""
        if not isinstance(list_of_points, list):
            return False
        try:
            point = lambda p: (
                (len(p) == 2) and self.check_if_numeric(p[0]) and self.check_if_numeric(p[1])
            )
            return all(map(point, list_of_points))
        except (KeyError, TypeError):
            return False


class BezierCalculator:
    @staticmethod
    def binomial(n: int, k: int):
        """Returns the binomial coefficient "n choose k" """
        return math.factorial(n) / float(math.factorial(k) * math.factorial(n - k))

    @staticmethod
    def bernstein_polynomial_point(x: int, i: int, n: int):
        """Calculate the i-th component of a bernstein polynomial of degree n"""
        return BezierCalculator.binomial(n, i) * (x**i) * ((1 - x) ** (n - i))

    @staticmethod
    def bernstein_polynomial(points: list):
        """
        Given list of control points, returns a function, which given a point [0,1] returns
        a point in the Bezier described by these points
        """

        def bernstein(t):
            n = len(points) - 1
            x = y = 0
            for i, point in enumerate(points):
                bern = BezierCalculator.bernstein_polynomial_point(t, i, n)
                x += point[0] * bern
                y += point[1] * bern
            return x, y

        return bernstein

    @staticmethod
    def calculate_points_in_curve(n: int, points: list):
        """
        Given list of control points, returns n points in the Bézier curve,
        described by these points
        """
        curvePoints = []
        bernstein_polynomial = BezierCalculator.bernstein_polynomial(points)
        for i in range(n):
            t = i / (n - 1)
            curvePoints += (bernstein_polynomial(t),)
        return curvePoints


class MouseMock:
    def __init__(self, page: Page):
        self.page = page
        self.last_x = 0
        self.last_y = 0

    async def click(
        self,
        x: int | float,
        y: int | float,
        button: Optional[str] = "left",
        click_count: Optional[int] = 1,
        delay: Optional[int] = 20,
        humanly: Optional[bool] = True,
    ) -> None:
        # Move mouse humanly to the Coordinates and wait some random time
        await self.move(x, y, humanly)
        await self.page.wait_for_timeout(random.randint(4, 8) * 100)

        # Clicking the Coordinates
        await self.page.mouse.down(button=button, click_count=click_count)
        # Waiting as delay
        await self.page.wait_for_timeout(delay)
        await self.page.mouse.up(button=button, click_count=click_count)

        # Waiting random time
        await self.page.wait_for_timeout(random.randint(4, 8) * 100)

    async def dblclick(
        self,
        x: int | float,
        y: int | float,
        button: Optional[str] = "left",
        delay: Optional[int] = 20,
        humanly: Optional[bool] = True,
    ) -> None:
        # Move mouse humanly to the Coordinates and wait some random time
        await self.move(x, y, humanly)
        await self.page.wait_for_timeout(random.randint(4, 8) * 100)

        # Clicking the Coordinates
        await self.page.mouse.down(button=button)
        # Waiting as delay
        await self.page.wait_for_timeout(delay)
        await self.page.mouse.up(button=button)

        # Waiting short random time
        await self.page.wait_for_timeout(random.randint(8, 14) * 10)
        # Clicking the Coordinates
        await self.page.mouse.down(button=button)
        # Waiting as delay
        await self.page.wait_for_timeout(delay)
        await self.page.mouse.up(button=button)

        # Waiting random time
        await self.page.wait_for_timeout(random.randint(4, 8) * 100)

    async def move(
        self,
        x: int | float,
        y: int | float,
        steps: Optional[int] = 1,
        humanly: Optional[bool] = True,
    ) -> None:
        # If you want to move in a straight line
        if not humanly:
            await self.page.mouse.origin_move(x, y, steps=steps)
            return

        if x == self.last_x and y == self.last_y:
            await self.page.wait_for_timeout(random.randint(1, 10))
            return

        humanized_points = HumanizeMouseTrajectory((self.last_x, self.last_y), (x, y))

        # Move Mouse to new random locations
        for x, y in humanized_points.points:
            await self.page.mouse.origin_move(x, y)
            # await page.wait_for_timeout(random.randint(1, 5))
        # Set LastX and LastY cause Playwright does not have mouse.current_location
        self.last_x, self.last_y = humanized_points.points[-1]



class ClientException(IOError):
    '''Error with the TLS client'''


class BrowserException(Exception):
    '''Base exceptions for render instances'''


class EnableMockHumanException(BrowserException):
    '''Exception raised when mock human is disabled, but captcha is called'''


class BrowserTimeoutException(BrowserException):
    '''Exception raised when playwright throws a timeout error'''


class NotRenderedException(Exception):
    '''Raise when the user tries to interact with an element that is not in a BrowserSession'''


class JavascriptException(BrowserException):
    '''Exception raised when a javascript error occurs'''


class SelectorNotFoundException(Exception):
    '''Exception raised when a css selector is not found'''


from . import page


async def new_context(inst, proxy, faker, mock_human, user_agent, **launch_arguments):
    # Spawning a new Context for more options
    context = await inst.main_browser.new_context(
        locale='en-US',
        geolocation={'longitude': proxy.longitude, 'latitude': proxy.latitude, 'accuracy': 0.7},
        timezone_id=proxy.timezone,
        permissions=['geolocation'],
        screen={'width': faker.avail_width, 'height': faker.avail_height},
        user_agent=user_agent,
        viewport={'width': faker.width, 'height': faker.height},
        proxy=proxy.browser_proxy,
        http_credentials={'username': proxy.username, 'password': proxy.password}
        if proxy.username
        else None,
        **launch_arguments,
    )

    if mock_human:
        await mock_context(inst, context, proxy, faker)
    return context


async def mock_context(inst, context, proxy, faker) -> None:
    async def page_mocker(**launch_arguments):
        return await page.new_page(inst, context, proxy, faker, **launch_arguments)

    context._new_page = context.new_page
    context.new_page = page_mocker



import asyncio
import hashlib
import logging
import os
from base64 import b64decode
from dataclasses import dataclass
from os.path import exists, isdir, join
from typing import Iterable, List, Union

import orjson
from Crypto.PublicKey import RSA


@dataclass
class Extension:
    '''
    Dataclass for managing Chrome extensions

    Args:
        path (str): path to the extension folder
    '''

    path: str

    def __post_init__(self):
        # open manifest.json
        manifest_path = join(self.path, 'manifest.json')
        assert exists(manifest_path), f'Extension "{self.path}" does not have a manifest.json file'
        # read manifest.json
        with open(manifest_path, 'rb') as f:
            manifest = orjson.loads(f.read())
        # check if manifest is mv3
        self.is_mv3: bool = manifest['manifest_version'] == 3
        # if the manifest has a 'key' field, use it to generate the extension id
        if 'key' in manifest:
            self.id: str = self.build_id(manifest['key'])
        # else, generate a random pub key to use instead
        else:
            manifest['key'] = self.generate_pub_key()
            self.id: str = self.build_id(manifest['key'])
            # write the new manifest.json
            with open(manifest_path, 'wb') as f:
                f.write(orjson.dumps(manifest, option=orjson.OPT_INDENT_2))
            logging.info(f'Generated new extension keypair ({self.id}) for "{self.path}"')

        self.url: str = f'chrome://extensions/?id={self.id}'

    @staticmethod
    def generate_pub_key() -> str:
        '''generate an RSA key and export it as a PKCS8 public key'''
        key = RSA.generate(2048)
        export = key.publickey().export_key(pkcs=8).decode('utf-8')
        return ''.join(export.split('\n')[1:-1])

    @staticmethod
    def build_id(pub_key_pem) -> str:
        '''build an extension id from a public key'''
        pub_key_der = b64decode(pub_key_pem)
        sha = hashlib.sha256(pub_key_der).hexdigest()
        prefix = sha[:32]
        return ''.join(chr(97 + int(old_char, 16)) for old_char in prefix)


class BuildExtensions:
    def __init__(self, extensions: Union[str, Iterable[str]]):
        self.list: List[Extension]

        # if `extension` is a string, assume it is a path to a folder of unpacked extensions
        if isinstance(extensions, str):
            assert isdir(extensions), 'Extension path must be a folder'
            # create a list of all the folders in the extension path
            self.list = []
            for item in os.listdir(extensions):
                path = join(extensions, item)
                if isdir(path):
                    self.list.append(Extension(path))
        # if `extension` is a list, assume it is a list of paths to unpacked extensions
        elif isinstance(extensions, list):
            assert all(isdir(path) for path in extensions), 'Extensions must be folders'
            self.list = [Extension(path) for path in extensions]


async def load_chrome_exts(page, exts: List[Extension]):
    for ext in exts:
        await page.goto(ext.url)
        await page.evaluate(
            "document.querySelector('extensions-manager').shadowRoot"
            ".querySelector('#viewManager > extensions-detail-view.active').shadowRoot"
            ".querySelector('div#container.page-container > div.page-content > div#options-section extensions-toggle-row#allow-incognito').shadowRoot"
            ".querySelector('label#label input').click()"
        )
    await page.goto('about:blank')


class LoadFirefoxAddon:
    '''
    Firefox addon loader
    Based on this Node.js implementation:
    https://github.com/microsoft/playwright/issues/7297#issuecomment-1211763085
    '''

    def __init__(self, port, addon_path):
        self.port: int = port
        self.addon_path: str = addon_path
        self.success: bool = False
        self.buffers: list = []
        self.remaining_bytes: int = 0

    async def load(self):
        reader, writer = await asyncio.open_connection('localhost', self.port)
        writer.write(self._format_message({"to": "root", "type": "getRoot"}))
        await writer.drain()

        while True:
            data = await reader.read(100)  # Adjust buffer size as needed
            if not data:
                break
            await self._process_data(writer, data)

        writer.close()
        await writer.wait_closed()
        return self.success

    async def _process_data(self, writer, data):
        while data:
            if self.remaining_bytes == 0:
                index = data.find(b':')
                if index == -1:
                    self.buffers.append(data)
                    return

                total_data = b''.join(self.buffers) + data
                size, _, remainder = total_data.partition(b':')

                try:
                    self.remaining_bytes = int(size)
                except ValueError as e:
                    raise ValueError("Invalid state") from e

                data = remainder

            if len(data) < self.remaining_bytes:
                self.remaining_bytes -= len(data)
                self.buffers.append(data)
                return
            else:
                self.buffers.append(data[: self.remaining_bytes])
                message = orjson.loads(b''.join(self.buffers))
                self.buffers.clear()

                await self._on_message(writer, message)

                data = data[self.remaining_bytes :]
                self.remaining_bytes = 0

    async def _on_message(self, writer, message):
        if "addonsActor" in message:
            writer.write(
                self._format_message(
                    {
                        "to": message["addonsActor"],
                        "type": "installTemporaryAddon",
                        "addonPath": self.addon_path,
                    }
                )
            )
            await writer.drain()

        if "addon" in message:
            self.success = True
            writer.write_eof()

        if "error" in message:
            writer.write_eof()

    def _format_message(self, data):
        raw = orjson.dumps(data)
        return f"{len(raw)}:".encode() + raw



import typing

from . import element_handle, frame_locator, js_handle, locator


def mock_frame(frame) -> None:
    if not frame:
        return

    async def click_mocker(
        selector,
        button="left",
        click_count=1,
        strict=False,
        delay=20,
        force=False,
        modifiers=[],
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await Frame.click(
            frame,
            selector,
            button=button,
            click_count=click_count,
            strict=strict,
            delay=delay,
            force=force,
            modifiers=modifiers,
            no_wait_after=no_wait_after,
            position=position,
            timeout=timeout,
            trial=trial,
        )

    frame.click = click_mocker

    async def dblclick_mocker(
        selector,
        button="left",
        strict=False,
        delay=20,
        force=False,
        modifiers=[],
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await Frame.dblclick(
            frame,
            selector,
            button=button,
            strict=strict,
            delay=delay,
            force=force,
            modifiers=modifiers,
            no_wait_after=no_wait_after,
            position=position,
            timeout=timeout,
            trial=trial,
        )

    frame.dblclick = dblclick_mocker

    async def check_mocker(
        selector,
        force=False,
        no_wait_after=False,
        position={},
        strict=False,
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await Frame.check(
            frame,
            selector,
            force=force,
            no_wait_after=no_wait_after,
            position=position,
            strict=strict,
            timeout=timeout,
            trial=trial,
        )

    frame.check = check_mocker

    async def uncheck_mocker(
        selector,
        force=False,
        no_wait_after=False,
        position={},
        strict=False,
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await Frame.uncheck(
            frame,
            selector,
            force=force,
            no_wait_after=no_wait_after,
            position=position,
            strict=strict,
            timeout=timeout,
            trial=trial,
        )

    frame.uncheck = uncheck_mocker

    async def set_checked_mocker(
        selector,
        checked=False,
        force=False,
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await Frame.set_checked(
            frame,
            selector,
            checked=checked,
            force=force,
            no_wait_after=no_wait_after,
            position=position,
            timeout=timeout,
            trial=trial,
        )

    frame.set_checked = set_checked_mocker

    async def hover_mocker(
        selector,
        force=False,
        modifiers=[],
        position={},
        strict=False,
        timeout: typing.Optional[float] = None,
        trial=False,
    ):
        await Frame.hover(
            frame,
            selector,
            force=force,
            modifiers=modifiers,
            position=position,
            strict=strict,
            timeout=timeout,
            trial=trial,
        )

    frame.hover = hover_mocker

    async def type_mocker(
        selector,
        text,
        delay=200,
        no_wait_after=False,
        strict=False,
        timeout: typing.Optional[float] = None,
    ):
        await Frame.type(
            frame,
            selector,
            text,
            delay=delay,
            no_wait_after=no_wait_after,
            strict=strict,
            timeout=timeout,
        )

    frame.type = type_mocker

    for frame in frame.child_frames:
        mock_frame(frame)

    # ElementHandle
    async def mock_query_selector(selector, strict=False) -> typing.Optional["ElementHandle"]:
        element = await frame._query_selector(selector, strict=strict)
        if element:
            await element_handle.mock_element_handle(element, frame.page)
        return element

    frame._query_selector = frame.query_selector
    frame.query_selector = mock_query_selector

    async def mock_query_selector_all(selector) -> typing.List["ElementHandle"]:
        elements = await frame._query_selector_all(selector)
        for element in elements:
            await element_handle.mock_element_handle(element, frame.page)
        return elements

    frame._query_selector_all = frame.query_selector_all
    frame.query_selector_all = mock_query_selector_all

    async def mock_wait_for_selector(
        selector, state=[], strict=False, timeout: typing.Optional[float] = None
    ) -> typing.Optional["ElementHandle"]:
        element = await frame.__wait_for_selector(
            selector, state=state, strict=strict, timeout=timeout
        )
        if element:
            await element_handle.mock_element_handle(element, frame.page)
        return element

    frame.__wait_for_selector = frame.wait_for_selector
    frame.wait_for_selector = mock_wait_for_selector

    async def mock_add_script_tag(content="", path="", type="", url="") -> "ElementHandle":
        element = await frame._add_script_tag(content=content, path=path, type=type, url=url)
        await element_handle.mock_element_handle(element, frame.page)
        return element

    frame._add_script_tag = frame.add_script_tag
    frame.add_script_tag = mock_add_script_tag

    async def mock_add_style_tag(content="", path="", type="", url="") -> "ElementHandle":
        element = await frame._add_script_tag(content=content, path=path, type=type, url=url)
        await element_handle.mock_element_handle(element, frame.page)
        return element

    frame._add_script_tag = frame.add_script_tag
    frame.add_script_tag = mock_add_style_tag

    async def frame_element() -> "ElementHandle":
        element = await frame._frame_element()
        await element_handle.mock_element_handle(element, frame.page)
        return element

    frame._frame_element = frame.frame_element
    frame.frame_element = frame_element

    # JsHandle
    async def mock_evaluate_handle(expression, arg=None) -> "JSHandle":
        _js_handle = await frame._evaluate_handle(expression, arg=arg)
        await js_handle.mock_js_handle(_js_handle, frame.page)
        return _js_handle

    frame._evaluate_handle = frame.evaluate_handle
    frame.evaluate_handle = mock_evaluate_handle

    async def mock_wait_for_function(
        expression, arg=None, polling="raf", timeout: typing.Optional[float] = None
    ) -> "JSHandle":
        _js_handle = await frame._wait_for_function(
            expression, arg=arg, polling=polling, timeout=timeout
        )
        await js_handle.mock_js_handle(_js_handle, frame.page)
        return _js_handle

    frame._wait_for_function = frame.wait_for_function
    frame.wait_for_function = mock_wait_for_function

    # FrameLocator
    async def frame_locator_mocker(selector) -> "JSHandle":
        _frame_locator = await frame._frame_locator(selector)
        await frame_locator.mock_frame_locator(_frame_locator)
        return _frame_locator

    frame._frame_locator = frame.frame_locator
    frame.frame_locator = frame_locator_mocker

    # Locator
    async def locator_mocker(selector, has=None, has_text="") -> "JSHandle":
        _locator = await frame._locator(selector, has=has, has_text=has_text)
        locator.mock_locator(_locator)
        return _locator

    frame._locator = frame.locator
    frame.locator = locator_mocker


class Frame:
    async def click(
        frame,
        selector,
        button="left",
        click_count=1,
        strict=False,
        delay=20,
        force=False,
        modifiers=[],
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        element = await frame.wait_for_selector(
            selector, state="visible" if not force else "hidden", strict=strict, timeout=timeout
        )

        if not force:
            await element.wait_for_element_state("editable", timeout=timeout)

        if not trial:
            if frame.page.scroll_into_view:
                await locator.scroll_into_view_if_needed(timeout=timeout)

            boundings = await element.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            for modifier in modifiers:
                await frame.page.keyboard.down(modifier)

            await frame.page.mouse.click(x, y, button, click_count, delay)

            for modifier in modifiers:
                await frame.page.keyboard.up(modifier)

    async def dblclick(
        frame,
        selector,
        button="left",
        strict=False,
        delay=20,
        force=False,
        modifiers=[],
        no_wait_after=False,
        position={},
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        element = await frame.wait_for_selector(
            selector, state="visible" if not force else "hidden", strict=strict, timeout=timeout
        )

        if not force:
            await element.wait_for_element_state("editable", timeout=timeout)

        if not trial:
            if frame.page.scroll_into_view:
                await locator.scroll_into_view_if_needed(timeout=timeout)

            boundings = await element.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            for modifier in modifiers:
                await frame.page.keyboard.down(modifier)

            await frame.page.mouse.dblclick(x, y, button, delay)

            for modifier in modifiers:
                await frame.page.keyboard.up(modifier)

    async def check(
        frame,
        selector,
        force=False,
        no_wait_after=False,
        position={},
        strict=False,
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        element = await frame.wait_for_selector(
            selector, state="visible" if not force else "hidden", strict=strict, timeout=timeout
        )

        if not force:
            await element.wait_for_element_state("editable", timeout=timeout)

        if await element.is_checked():
            return

        if not trial:
            if frame.page.scroll_into_view:
                await locator.scroll_into_view_if_needed(timeout=timeout)

            boundings = await element.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            await frame.page.mouse.click(x, y, button="left", click_count=1, delay=20)

            assert await element.is_checked()

    async def uncheck(
        frame,
        selector,
        force=False,
        no_wait_after=False,
        position={},
        strict=False,
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        element = await frame.wait_for_selector(
            selector, state="visible" if not force else "hidden", strict=strict, timeout=timeout
        )

        if not force:
            await element.wait_for_element_state("editable", timeout=timeout)

        if not await element.is_checked():
            return

        if not trial:
            if frame.page.scroll_into_view:
                await locator.scroll_into_view_if_needed(timeout=timeout)

            boundings = await element.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            await frame.page.mouse.click(x, y, button="left", click_count=1, delay=20)

            assert not await element.is_checked()

    async def set_checked(
        frame,
        selector,
        checked=False,
        force=False,
        no_wait_after=False,
        position={},
        strict=False,
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        element = await frame.wait_for_selector(
            selector, state="visible" if not force else "hidden", strict=strict, timeout=timeout
        )

        if not force:
            await element.wait_for_element_state("editable", timeout=timeout)

        if await element.is_checked() == checked:
            return

        if not trial:
            if frame.page.scroll_into_view:
                await locator.scroll_into_view_if_needed(timeout=timeout)

            boundings = await element.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            await frame.page.mouse.click(x, y, button="left", click_count=1, delay=20)

            assert await element.is_checked()

    async def hover(
        frame,
        selector,
        force=False,
        modifiers=[],
        position={},
        strict=False,
        timeout: typing.Optional[float] = None,
        trial=False,
    ) -> None:
        element = await frame.wait_for_selector(
            selector, state="visible" if not force else "hidden", strict=strict, timeout=timeout
        )

        if not force:
            await element.wait_for_element_state("editable", timeout=timeout)

        if not trial:
            if frame.page.scroll_into_view:
                await locator.scroll_into_view_if_needed(timeout=timeout)

            boundings = await element.bounding_box()
            x, y, width, height = boundings.values()
            if not position:
                x, y = x + width // 2, y + height // 2
            else:
                x, y = x + position["x"], y + position["y"]

            for modifier in modifiers:
                await frame.page.keyboard.down(modifier)

            await frame.page.mouse.move(x, y)

            for modifier in modifiers:
                await frame.page.keyboard.up(modifier)

    async def type(
        frame,
        selector,
        text,
        delay=200,
        no_wait_after=False,
        strict=False,
        timeout: typing.Optional[float] = None,
    ) -> None:
        element = await frame.wait_for_selector(
            selector, state="visible", strict=strict, timeout=timeout
        )

        await element.wait_for_element_state("editable", timeout=timeout)

        if frame.page.scroll_into_view:
            await locator.scroll_into_view_if_needed(timeout=timeout)

        boundings = await element.bounding_box()
        x, y, width, height = boundings.values()

        x, y = x + width // 2, y + height // 2

        await frame.page.mouse.click(x, y, "left", 1, delay)

        await frame.page.keyboard.type(text, delay=delay)



import traceback
from concurrent.futures import Future, ThreadPoolExecutor
from concurrent.futures import wait as futures_wait
from dataclasses import dataclass
from functools import partial
from threading import Thread
from typing import Callable, Dict, Iterable, List, Optional, Union, overload
from urllib.parse import urlencode

import gevent
from gevent.pool import Pool

import playfulrequests
from playfulrequests.response import Response


class TLSRequest:
    '''
    Asynchronous request.
    Accepts the same parameters as ``TLSSession.request`` and some additional parameters:

    Args:
        method (str): Request method.
        url (str): URL to request.
        session (playfulrequests.session.TLSSession, optional): Associated `TLSSession`. Defaults to None.
        raise_exception (bool, optional): Raise exceptions (default FALSE for async, TRUE for sync). Defaults to False.
        params (Dict[str, str], optional): Parameters to pass to request. Defaults to None.
        callback (function, optional): Callback called on response. Same as passing ``hooks={'response': callback}``. Defaults to None.

    Attributes:
        method (str): Request method.
        raise_exception (bool): Raise exceptions (default FALSE for async, TRUE for sync).
        url (str): URL to request.
        session (playfulrequest.session.TLSSession): Associated `TLSSession`.
        kwargs (dict): The rest arguments for ``Session.request``.
        response (playfulrequest.response.Response): Resulting ``Response``.

    Methods:
        send(**kwargs): Prepares request based on parameter passed to constructor and optional ``kwargs```.
                        Then sends request and saves response to :attr:`response`.
                        Returns: ``Response``.
    '''

    session_kwargs = {
        'browser',
        'version',
        'os',
        'ja3_string',
        'h2_settings',
        'additional_decode',
        'pseudo_header_order',
        'priority_frames',
        'header_order',
        'force_http1',
        'catch_panics',
        'debug',
    }

    def __init__(
        self,
        method: str,
        url: str,
        session: Optional['playfulrequests.session.TLSSession'] = None,
        params: Optional[Dict[str, str]] = None,
        raise_exception: bool = True,
        **kwargs,
    ):
        # Request method
        self.method: str = method
        # Raise exceptions (default FALSE for async, TRUE for sync)
        self.raise_exception: bool = raise_exception
        # URL to request
        if params is None:
            self.url: str = url
        else:
            self.url: str = f'{url}?{urlencode(params, doseq=True)}'

        # Session kwargs
        self.sess_kwargs: Optional[dict] = None
        if kwargs:
            sess_kwargs = set(kwargs.keys()) & TLSRequest.session_kwargs
            if session and sess_kwargs:
                # If session is already provided, raise TypeError if session-only kwargs are passed
                raise TypeError(f'Cannot pass parameter(s) to an existing session: {sess_kwargs}')
            else:
                self.sess_kwargs = {k: kwargs.pop(k) for k in sess_kwargs}

        if callback := kwargs.pop('callback', None):
            kwargs['hooks'] = {'response': callback}

        # The rest of the arguments for `Session.request`
        self.kwargs = kwargs
        # Resulting Response
        self.response = None
        # Create TLSSession if not provided
        self._build_session(session)

    def _build_session(self, session=None):
        if session is None:
            if self.sess_kwargs:
                # if session kwargs are passed, configure a new session with them
                self.session = playfulrequests.Session(temp=True, **self.sess_kwargs)
            else:
                # else use a preconfigured session
                self.session = playfulrequests.chrome.Session(temp=True)
            self._close = True
        else:
            # don't close adapters after each request if the user provided the session
            self.session = session
            self._close = False

    def send(self, **kwargs):
        '''
        Prepares request based on parameter passed to constructor and optional ``kwargs```.
        Then sends request and saves response to :attr:`response`
        :returns: ``Response``
        '''
        merged_kwargs = {}
        merged_kwargs.update(self.kwargs)
        merged_kwargs.update(kwargs)
        # rebuild session if it was deleted
        if self.session is None:
            self._build_session()
        try:
            self.response = self.session.request(self.method, self.url, **merged_kwargs)
        except Exception as e:
            if self.raise_exception:
                raise e
            self.exception = e
            self.traceback = traceback.format_exc()
        finally:
            self.close_session()
        return self

    def close_session(self) -> None:
        if self._close and self.session is not None:
            # close the session if it was created by this request
            self.session.close()
            self.session = None


class LazyTLSRequest(TLSRequest):
    '''
    This will send the request immediately, but doesn't wait for the response to be ready
    until an attribute of the response is accessed
    '''

    def __init__(self, *args, **kwargs):
        executor: Optional[ThreadPoolExecutor] = kwargs.pop('executor', None)
        super().__init__(*args, **kwargs)

        self._thread: Union[Thread, Future]
        if executor:
            self._thread = executor.submit(self._send)
        else:
            self._thread = Thread(target=self._send)
            self._thread.start()
        self.complete: bool = False

    def __repr__(self):
        return self.response.__repr__() if self.complete else '<LazyResponse[Pending]>'

    def _send(self):
        super().send()
        self.complete = True

    def join(self):
        if isinstance(self._thread, Future):
            # await future to be ready
            futures_wait((self._thread,))
        else:
            # await thread to be ready
            self._thread.join()

    def __getattr__(self, name: str):
        # if an attribute is called, JOIN the greenlet and continue
        if not self.complete:
            self.join()
        return getattr(self.response, name)


def request_list(
    method: str, url: Iterable[str], *args, **kwargs
) -> Iterable[Union[Response, LazyTLSRequest]]:
    '''
    Concurrently send requests given a list of urls
    '''
    # if wait is False, return a tuple of LazyTLSRequests
    if kwargs.pop('nohup', None):
        executor = ThreadPoolExecutor()
        # return a list of LazyTLSRequests objs
        return [
            LazyTLSRequest(method, u, *args, **kwargs, executor=executor, raise_exception=False)
            for u in url
        ]
    # send requests to urls concurrently with map
    return map([async_request(method, u, *args, **kwargs) for u in url])


@overload
def request(
    method: str, url: Iterable[str], *args, **kwargs
) -> Iterable[Union[Response, LazyTLSRequest]]:
    ...


@overload
def request(method: str, url: str, *args, **kwargs) -> Union[Response, LazyTLSRequest]:
    ...


def request(method: str, url: Union[str, Iterable[str]], *args, **kwargs):
    '''
    Send a request with TLS client

    Args:
        method (str): Method of request (GET, POST, OPTIONS, HEAD, PUT, PATCH, DELETE)
        url (Union[str, Iterable[str]]): URL or list of URLs to request.
        params (dict, optional): Dictionary of URL parameters to append to the URL. Defaults to None.
        data (Union[str, dict], optional): Data to send to request. Defaults to None.
        headers (dict, optional): Dictionary of HTTP headers to send with the request. Defaults to None.
        cookies (Union[dict, RequestsCookieJar], optional): Dict or CookieJar to send. Defaults to None.
        json (dict, optional): Json to send in the request body. Defaults to None.
        allow_redirects (bool, optional): Allow request to redirect. Defaults to True.
        verify (bool, optional): Verify the server's TLS certificate. Defaults to True.
        timeout (int, optional): Timeout in seconds. Defaults to 30.
        proxies (dict, optional): Dictionary of proxies. Defaults to None.
        wait (bool, optional): Wait for response to be ready. Defaults to True.
        threadsafe (bool, optional): Threadsafe support for wait=False. Defaults to False.

    Returns:
        playfulrequest.response.Response: Response object
    '''
    # if a list of urls is passed, send requests concurrently
    if isinstance(url, (list, tuple)):
        return request_list(method, url, *args, **kwargs)
    # if nohup is True, return a LazyTLSRequest
    if kwargs.pop('nohup', None):
        return LazyTLSRequest(method, url, *args, **kwargs)
    req = TLSRequest(method, url, *args, **kwargs)
    req.send()
    return req.response


def async_request(*args, raise_exception=False, **kwargs) -> TLSRequest:
    '''
    Return an unsent request to be used with map, imap, and imap_enum.
    Used to send requests concurrently.

    Args:
        method (str): Method of request (GET, POST, OPTIONS, HEAD, PUT, PATCH, DELETE)
        url (str): URL to send request to
        params (dict, optional): Dictionary of URL parameters to append to the URL. Defaults to None.
        data (Union[str, dict], optional): Data to send to request. Defaults to None.
        headers (dict, optional): Dictionary of HTTP headers to send with the request. Defaults to None.
        cookies (Union[dict, RequestsCookieJar], optional): Dict or CookieJar to send. Defaults to None.
        json (dict, optional): Json to send in the request body. Defaults to None.
        allow_redirects (bool, optional): Allow request to redirect. Defaults to True.
        verify (bool, optional): Verify the server's TLS certificate. Defaults to True.
        timeout (int, optional): Timeout in seconds. Defaults to 30.
        proxies (dict, optional): Dictionary of proxies. Defaults to None.

    Returns:
        TLSRequest: Unsent request object
    '''
    return TLSRequest(*args, **kwargs, raise_exception=raise_exception)


'''
Requests shortcuts
'''

# Shortcuts for creating synchronous requests
get: partial = partial(request, 'GET')
options: partial = partial(request, 'OPTIONS')
head: partial = partial(request, 'HEAD')
post: partial = partial(request, 'POST')
put: partial = partial(request, 'PUT')
patch: partial = partial(request, 'PATCH')
delete: partial = partial(request, 'DELETE')

'''
Asynchronous requests shortcuts
'''

# Shortcuts for creating an unsent TLSRequest
async_get: partial = partial(async_request, 'GET')
async_post: partial = partial(async_request, 'POST')
async_options: partial = partial(async_request, 'OPTIONS')
async_head: partial = partial(async_request, 'HEAD')
async_put: partial = partial(async_request, 'PUT')
async_patch: partial = partial(async_request, 'PATCH')
async_delete: partial = partial(async_request, 'DELETE')


def send(r, pool: Optional[Pool] = None):
    '''
    Sends the request object using the specified pool. If a pool isn't
    specified this method blocks. Pools are useful because you can specify size
    and can hence limit concurrency.
    '''
    return gevent.spawn(r.send) if pool is None else pool.spawn(r.send)


@dataclass
class FailedResponse:
    '''
    A FailedResponse object is returned when a request fails and no exception handler is provided.
    '''

    exception: Exception

    def __bool__(self) -> bool:
        return False

    def __repr__(self) -> str:
        return f'<FailedResponse: {self.exception}>'

    def __str__(self) -> str:
        return str(self.exception)


def map(
    requests: List[TLSRequest],
    size: Optional[int] = None,
    exception_handler: Optional[Callable] = None,
):
    '''
    Concurrently converts a list of Requests to Responses.

    Parameters:
        requests - a collection of Request objects.
        size - Specifies the number of requests to make at a time. If None, no throttling occurs.
        exception_handler - Callback function, called when exception occurred. Params: Request, Exception

    Returns:
        A list of Response objects.
    '''

    requests = list(requests)
    all_resps: List[Optional[Response]] = []

    if size is None:
        # set default increment size to the total
        size = len(requests)

    for inc in range(0, len(requests), size):
        processed_reqs: List[playfulrequests.response.ProcessResponse] = []
        requests_range = requests[inc : min(inc + size, len(requests))]
        for req in requests_range:
            # prepare the request & construct sessions
            if req.session is None:
                req._build_session()
            # create a list of ProcessResponse objects
            processed_reqs.append(
                req.session.request(req.method, req.url, **req.kwargs, process=False)
            )
        try:
            resps: List[Optional[Response]] = playfulrequests.response.ProcessResponsePool(
                processed_reqs
            ).execute_pool()
        except Exception as e:
            # handle exception for all requests in the pool
            failed_resp: FailedResponse = FailedResponse(e)  # create a FailedResponse object
            resps = [failed_resp] * len(requests_range)  # add None for each failed request
            for req in requests_range:
                if req.raise_exception:
                    raise e
                req.exception = e
                req.traceback = traceback.format_exc()
                if exception_handler:
                    exception_handler(req, e)
        finally:
            # close sessions
            for req in requests_range:
                req.close_session()
        all_resps.extend(resps)
    return all_resps


def imap(
    requests: List[TLSRequest],
    size: int = 2,
    enumerate: bool = False,
    exception_handler: Optional[Callable] = None,
):
    '''
    Concurrently converts a generator object of Requests to a generator of Responses.

    Parameters:
        requests - a generator or sequence of Request objects.
        size - Specifies the number of requests to make at a time. default is 2
        exception_handler - Callback function, called when exception occurred. Params: Request, Exception

    Yields:
        Response objects.
    '''
    if enumerate:  # send to imap_enum
        return imap_enum(requests, size, exception_handler)

    def send(r):
        return r.send()

    pool = Pool(size)
    for request in pool.imap_unordered(send, requests):
        if request.response is not None:
            yield request.response
        elif exception_handler:
            ex_result = exception_handler(request, request.exception)
            if ex_result is not None:
                yield ex_result
        else:
            yield FailedResponse(request.exception)

    pool.join()


def imap_enum(
    requests: List[TLSRequest],
    size: int = 2,
    exception_handler: Optional[Callable] = None,
):
    '''
    Like imap, but yields tuple of original request index and response object
    Unlike imap, failed results and responses from exception handlers that return None are not ignored. Instead, a
    tuple of (index, None) is yielded.
    Responses are still in arbitrary order.

    Parameters:
        requests - a sequence of Request objects.
        size - Specifies the number of requests to make at a time. default is 2
        exception_handler - Callback function, called when exception occurred. Params: Request, Exception

    Yields:
        (index, Response) tuples.
    '''

    def send(r):
        return r._index, r.send()

    requests = list(requests)
    for index, req in enumerate(requests):
        req._index = index

    pool = Pool(size)
    for index, request in pool.imap_unordered(send, requests):
        if request.response is not None:
            yield index, request.response
        elif exception_handler:
            ex_result = exception_handler(request, request.exception)
            yield index, ex_result
        else:
            yield index, FailedResponse(request.exception)



import asyncio
from functools import partial
from http.client import responses as status_codes
from threading import Thread
from typing import Any, Callable, Dict, Iterable, List, Literal, Optional, Pattern, Union

from aioprocessing import Queue
from playwright._impl._api_types import Error as PlaywrightError
from playwright._impl._api_types import TimeoutError as PlaywrightTimeoutError

import playfulrequests
from playfulrequests.client import CaseInsensitiveDict
from playfulrequests.cookies import cookiejar_to_list, list_to_cookiejar
from playfulrequests.exceptions import BrowserException, BrowserTimeoutException, JavascriptException
from playfulrequests.headers import Headers
from playfulrequests.response import Response

from .cookies import RequestsCookieJar
from .extensions import BuildExtensions, Extension, load_chrome_exts

_browsers = {
    'firefox': playfulrequests.FirefoxBrowser,
    'chrome': playfulrequests.ChromeBrowser,
}


class BrowserSession:
    """
    Args:
        headless (bool, optional): Whether to run the browser in headless mode. Defaults to True.
        session (playfulrequests.session.TLSSession, optional): Session to use for headers, cookies, etc.
        resp (playfulrequests.response.Response, optional): Response to update with cookies, headers, etc.
        proxy_ip (str, optional): Proxy to use for the browser. Example: 123.123.123
        mock_human (bool, optional): Whether to emulate human behavior. Defaults to False.
        browser (Literal['firefox', 'chrome'], optional): Generate useragent headers for a specific browser
        os (Literal['win', 'mac', 'lin'], optional): Generate headers for a specific OS
        extensions (Union[str, Iterable[str]], optional): Path to a folder of unpacked extensions, or a list of paths to unpacked extensions

    Attributes:
        url (str): Get the page url
        headers (dict): Get the browser headers (User-Agent)
        content (dict): Get the current page content
        cookies (RequestsCookieJar): Get the browser cookies
        status_code (int): Status code of the last response
        reason (Optional[str]): Gets the official W3C name for the status code

    Navigation Methods:
        goto(url): Navigate to a URL.
        forward(): Navigate to the next page in history
        back(): Navigate to the previous page in history
        awaitNavigation(): Wait for the page navigation to finish
        awaitScript(script, arg): Wait for a script to return true
        awaitSelector(selector, arg): Wait for a selector to exist
        awaitEnabled(selector, arg): Wait for a selector to be enabled
        isVisible(selector): Check if a selector is visible
        isEnabled(selector): Check if a selector is enabled
        awaitUrl(url, timeout): Wait for the URL to match
        dragTo(source, target, timeout, wait_after, check): Drag and drop a selector
        type(selector, text, delay, timeout): Type text into a selector
        click(selector, click_count, button, timeout, wait_after): Click a selector
        hover(selector, modifiers, timeout): Hover over a selector
        evaluate(script, arg): Evaluate and return a script
        screenshot(selector, path, full_page): Take a screenshot of the page
        setHeaders(headers): Set the browser headers. Note that this will NOT update the TLSSession headers
        close(): Close the instance

    Network Methods:
        get(url, params, headers, timeout, verify, max_redirects): Send a GET request
        post(url, params, headers, data, form, multipart, timeout, verify, max_redirects): Send a POST request
        put(url, params, headers, data, form, multipart, timeout, verify, max_redirects): Send a PUT request
        patch(url, params, headers, data, form, multipart, timeout, verify, max_redirects): Send a PATCH request
        delete(url, params, headers, timeout, verify, max_redirects): Send a DELETE request
        head(url, params, headers, timeout, verify, max_redirects): Send a HEAD request
    """

    def __init__(
        self,
        *,
        headless: bool = True,
        session: Optional[playfulrequests.session.TLSSession] = None,
        resp: Optional[playfulrequests.response.Response] = None,
        proxy_ip: Optional[str] = None,
        mock_human: bool = False,
        browser: Optional[Literal['firefox', 'chrome']] = None,
        os: Optional[Literal['win', 'mac', 'lin']] = None,
        extensions: Optional[Union[str, Iterable[str]]] = None,
    ) -> None:
        # uses asyncio queues to communicate with the asyncio loop from any thread
        # _in is for calls from other threads
        self._in: Queue = Queue()
        # _out is for responses from the asyncio loop
        self._out: Queue = Queue()
        # remember session and resp to clone cookies back to when closing
        self.session: Optional[playfulrequests.session.TLSSession] = session
        self.resp: Optional[playfulrequests.response.Response] = resp
        # use the passed browser
        self.browser: Literal['firefox', 'chrome'] = browser or 'firefox'
        # generating headers
        if session:
            # if a session was provided, use the session user-agent
            self.ua: str = session.headers.get('User-Agent')
        else:
            # if a browser or os was provided, generate a user-agent and IGNORE session/resp headers
            # only meant to be used when using BrowserSession as a standalone
            self.ua: str = Headers(browser=self.browser, os=os, headers=False).generate()[
                'User-Agent'
            ]
        # proxy variables
        self.proxy_ip: Optional[str] = proxy_ip
        # browser config
        self.status_code: Optional[int]
        if self.resp is not None:
            self.status_code = self.resp.status_code
        else:
            self.status_code = None
        self.mock_human: bool = mock_human
        self.headless: bool = headless
        self.extensions: Optional[List[Extension]] = (
            BuildExtensions(extensions).list if extensions else None
        )
        # bool to indicate browser was closed
        self._closed: bool = False
        # spawn asyncio loop
        thread: Thread = Thread(target=self.spawn_main, daemon=True)
        thread.start()

    def spawn_main(self) -> None:
        asyncio.new_event_loop().run_until_complete(self.main())

    async def main(self) -> None:
        # build the playwright instance
        self.client = await _browsers[self.browser](
            headless=self.headless, extensions=self.extensions
        )
        self.context = await self.client.new_context(
            browser_name=self.browser,
            user_agent=self.ua,
            proxy=self.proxy_ip,
            mock_human=self.mock_human,
        )
        # create a new page
        self.page = await self.context.new_page()
        # activate extensions on chrome
        if self.extensions and self.browser == 'chrome':
            await load_chrome_exts(self.page, self.extensions)
        '''
        run the main loop
        '''
        while True:
            try:
                # listen for calls to _in
                call: partial = self._in.get()
                # handle the call within the asyncio loop
                # this is used as a workaround to make playwright work across threads
                out = await call()
            except PlaywrightTimeoutError as e:
                # if a timeout error is raised, mark as closed and put the error in _out
                self._closed = True
                self._out.put(BrowserTimeoutException(e))
            except PlaywrightError as e:
                # if a playwright error is raised, mark as closed and put the error in _out
                self._closed = True
                self._out.put(BrowserException(e))
            except BrowserException as e:
                self._out.put(e)
            else:
                # put the call response in _out
                self._out.put(out)

    async def shutdown(self) -> None:
        self._closed = True
        await self.client.stop()

    def __enter__(self) -> 'BrowserSession':
        return self

    def __exit__(self, *_) -> None:
        self.close()

    def _call_wrapper(self, func, *args, **kwargs) -> Any:
        # the call will then be ran within the asyncio loop,
        # which will launch the corresponding function with a leading underscore
        # then return the result
        self._in.put(partial(func, *args, **kwargs))
        out = self._out.get()
        if isinstance(out, Exception):
            raise out
        return out

    def __getattr__(self, name) -> Any:
        # sourcery skip: raise-from-previous-error
        if self._closed:
            raise BrowserException(f'Browser was closed. Attribute call failed: {name}')
        # forwards unknown attribute calls to _call_wrapper
        try:
            # check if the attribute (with a leading _) exists
            attr = self.__getattribute__(f'_{name}')
        except AttributeError:
            # if it doesn't, raise an error
            raise AttributeError(
                "'{obj_name}' object has no attribute '{attr_name}'".format(
                    obj_name=self.__class__.__name__, attr_name=name
                )
            )
        return partial(self._call_wrapper, attr)

    """
    Common public functions
    """

    async def _goto(self, url):
        '''Navigate to a URL'''
        resp = await self.page.goto(url)
        self.status_code = resp.status
        return resp

    async def _forward(self):
        '''Navigate to the next page in history'''
        return await self.page.go_forward()

    async def _back(self):
        '''Navigate to the previous page in history'''
        return await self.page.go_back()

    async def _awaitNavigation(self, timeout: float = 30):
        '''
        Wait for the page navigation to finish

        Parameters:
            timeout (float, optional): Timeout in seconds. Defaults to 30.
        '''
        return await self.page.wait_for_load_state(timeout=int(timeout * 1e3))

    async def _awaitScript(self, script: str, arg: Optional[str] = None, *, timeout: float = 30):
        '''
        Wait for a script to return true

        Parameters:
            script (str): Script to evaluate
            arg (str, optional): Argument to pass to script
            timeout (float, optional): Timeout in seconds. Defaults to 30.
        '''
        return await self.page.wait_for_function(script, arg=arg, timeout=int(timeout * 1e3))

    async def _awaitSelector(self, selector, *, timeout: float = 30):
        '''
        Wait for a selector to exist

        Parameters:
            selector (str): Selector to wait for
            timeout (float, optional): Timeout in seconds. Defaults to 30.
        '''
        await self.page.wait_for_function(
            "selector => !!document.querySelector(selector)",
            arg=selector,
            timeout=int(timeout * 1e3),
        )

    async def _awaitEnabled(self, selector, *, timeout: float = 30):
        '''
        Wait for a selector to be enabled

        Parameters:
            selector (str): Selector to wait for
            timeout (float, optional): Timeout in seconds. Defaults to 30.
        '''
        await self.page.wait_for_function(
            "selector => !document.querySelector(selector).disabled",
            arg=selector,
            timeout=int(timeout * 1e3),
        )

    async def _isVisible(self, selector: str) -> bool:
        '''
        Check if a selector is visible

        Parameters:
            selector (str): Selector to check
        '''
        return await self.page.is_visible(selector)

    async def _isEnabled(self, selector: str) -> bool:
        '''
        Check if a selector is enabled

        Parameters:
            selector (str): Selector to check
        '''
        if not await self.page.is_visible(selector):
            return False
        return await self.page.evaluate(
            "selector => !document.querySelector(selector).disabled", arg=selector
        )

    async def _awaitUrl(
        self, url: Union[str, Pattern[str], Callable[[str], bool]], *, timeout: float = 30
    ):
        '''
        Wait for the url to match a string, regex, or a python function to return True

        Parameters:
            url (Union[str, Pattern[str], Callable[[str], bool]]) - URL to match for
            timeout (float, optional): Timeout in seconds. Defaults to 30.
        '''
        return await self.page.wait_for_url(url, timeout=int(timeout * 1e3))

    async def _dragTo(
        self,
        source: str,  # selector to drag on
        target: str,  # selector to drop on
        *,
        timeout: float = 30,  # timeout in seconds
        wait_after: bool = False,  # dont wait for a page event
        check: bool = False,  # check if element is able to be dragged
    ):
        '''
        Drag and drop a selector

        Parameters:
            source (str): Source to drag from
            target (str): Target to drop to
            timeout (float, optional): Timeout in seconds. Defaults to 30.
            wait_after (bool, optional): Wait for a page event before continuing. Defaults to False.
            check (bool, optional): Check if an element is draggable before running. Defaults to False.
        '''
        return await self.page.drag_and_drop(
            source, target, no_wait_after=not wait_after, timeout=int(timeout * 1e3), check=check
        )

    async def _type(self, selector: str, text: str, delay: int = 50, *, timeout: float = 30):
        '''
        Type text into a selector

        Parameters:
            selector (str): CSS selector to type in
            text (str): Text to type
            delay (int, optional): Delay between keypresses in ms. On mock_human, this is randomized by 50%. Defaults to 50.
            timeout (float, optional): Timeout in seconds. Defaults to 30.
        '''
        return await self.page.type(selector, text, delay=delay, timeout=int(timeout * 1e3))

    async def _click(
        self,
        selector: str,
        button: Optional[Literal['left', 'right', 'middle']] = 'left',
        count: int = 1,
        *,
        timeout: float = 30,
        wait_after: bool = True,
    ):
        '''
        Click a selector

        Parameters:
            selector (str): CSS selector to click
            button (Literal['left', 'right', 'middle'], optional): Button to click. Defaults to 'left'.
            count (int, optional): Number of clicks. Defaults to 1.
            timeout (float, optional): Timeout in seconds. Defaults to 30.
            wait_after (bool, optional): Wait for a page event before continuing. Defaults to True.
        '''
        return await self.page.click(
            selector,
            button=button,
            click_count=count,
            timeout=int(timeout * 1e3),
            no_wait_after=not wait_after,
        )

    async def _hover(
        self,
        selector: str,
        modifiers: Optional[List[Literal['Alt', 'Control', 'Meta', 'Shift']]] = None,
        *,
        timeout: float = 90,
    ):
        '''
        Hover over a selector

        Parameters:
            selector (str): CSS selector to hover over
            modifiers (List[Literal['Alt', 'Control', 'Meta', 'Shift']], optional): Modifier keys to press. Defaults to None.
            timeout (float, optional): Timeout in seconds. Defaults to 90.
        '''
        return await self.page.hover(selector, modifiers=modifiers, timeout=int(timeout * 1e3))

    async def _evaluate(self, script: str, arg: Optional[str] = None):
        '''
        Evaluate and return javascript

        Parameters:
            script (str): Javascript to evaluate in the page
            arg (str, optional): Argument to pass into the javascript function
        '''
        try:
            return await self.page.evaluate(script, arg=arg)
        except PlaywrightError as e:
            raise JavascriptException('Javascript eval exception') from e

    async def _screenshot(
        self, selector: Optional[str], path: Optional[str] = None, *, full_page: bool = False
    ) -> Optional[bytes]:
        '''
        Take a screenshot of the page

        Parameters:
            selector (str, optional): CSS selector to screenshot
            path (str, optional): Path to save screenshot to. Defaults to None.
            full_page (bool): Whether to take a screenshot of the full scrollable page. Cannot be used with selector. Defaults to False.
            
        Returns:
            Optional[bytes]: Returns the screenshot buffer, if `path` was not provided
        '''
        assert not (selector and full_page), 'Cannot use selector and full_page together'
        if selector:
            buffer = await self.page.locator(selector).screenshot(path=path)
        else:
            buffer = await self.page.screenshot(path=path, full_page=full_page)
        if not path:  # dont return buffer if path was provided
            return buffer

    '''
    .url, .content, .cookies, .html, properties
    makes this compatible with TLSSession
    '''

    async def _getContent(self):
        '''Get the page content'''
        return await self.page.content()

    async def _getCookies(self) -> RequestsCookieJar:
        '''Get the page cookies'''
        browser_cookies: list = await self.context.cookies()
        return list_to_cookiejar(browser_cookies)

    @property
    def url(self) -> str:
        '''Get the page url'''
        return self.page.url

    @url.setter
    def url(self, url: str):
        '''Go to page url'''
        self.goto(url)

    @property
    def headers(self) -> CaseInsensitiveDict:
        '''Get the page headers'''
        return CaseInsensitiveDict({'User-Agent': self.ua})

    @headers.setter
    def headers(self, headers: dict):
        '''Set headers'''
        self.setHeaders(headers)

    @property
    def content(self) -> str:
        '''Get the page url'''
        return self.getContent()

    @property
    def proxies(self):
        return {'all': self.proxy_ip} if self.proxy_ip else {}

    @proxies.setter
    def proxies(self, _: dict):
        raise NotImplementedError('Cannot set proxies on a browser session')

    @property
    def cookies(self) -> RequestsCookieJar:
        '''Get the context cookies'''
        return self.getCookies()

    @cookies.setter
    def cookies(self, cookiejar: RequestsCookieJar):
        '''Set the context cookies'''
        self.setCookies(cookiejar)

    @property
    def html(self) -> 'playfulrequests.parser.HTML':
        '''Get the page html as an HTML object'''
        return playfulrequests.parser.HTML(
            session=self, url=self.url, html=self.content
        )

    @property
    def find(self) -> Callable:
        return self.html.find
    
    @property
    def find_all(self) -> Callable:
        return self.html.find_all

    @property
    def reason(self) -> Optional[str]:
        if self.status_code is not None:
            return status_codes[self.status_code]

    '''
    Network request functions
    '''

    async def _request(
        self,
        method: Literal['GET', 'POST', 'PUT', 'DELETE', 'PATCH', 'HEAD'],
        url: str,
        *,
        params: Optional[Dict[str, Union[str, float, bool]]] = None,
        data: Optional[Union[Any, str, bytes]] = None,
        headers: Optional[Union[Dict[str, str], CaseInsensitiveDict]] = None,
        form: Optional[Dict[str, Union[str, float, bool]]] = None,
        multipart: Optional[Dict[str, Union[bytes, bool, float, str]]] = None,
        timeout: float = 30,
        verify: bool = True,
        max_redirects: Optional[int] = None,
    ) -> Response:
        '''
        Parameters:
            method (Literal['GET', 'POST', 'PUT', 'DELETE', 'PATCH', 'HEAD']): HTTP method to use
            url (str): URL to send request to
            params (dict, optional): Dictionary of URL parameters to append to the URL. Defaults to None.
            data (Union[str, dict], optional): Data to send to request. Defaults to None.
            headers (dict, optional): Dictionary of HTTP headers to send with the request. Defaults to None.
            form (dict, optional): Form data to send with the request. Defaults to None.
            multipart (dict, optional): Multipart data to send with the request. Defaults to None.
            timeout (float, optional): Timeout in seconds. Defaults to 30.
            verify (bool, optional): Verify the server's TLS certificate. Defaults to True.
            max_redirects (int, optional): Maximum number of redirects to follow. Defaults to None.
        '''
        # convert headers to dict if a CaseInsensitiveDict was provided
        if isinstance(headers, CaseInsensitiveDict):
            headers = dict(headers)
        pywr_resp = await self.context.request.fetch(
            url,
            params=params,
            method=method.lower(),
            headers=headers,
            data=data,
            form=form,
            multipart=multipart,
            timeout=int(timeout * 1e3),
            ignore_https_errors=not verify,
            max_redirects=max_redirects,
        )
        content = await pywr_resp.body()
        resp = Response(
            _content=content,
            url=pywr_resp.url,
            status_code=pywr_resp.status,
            cookies=await self._getCookies(),
            headers=CaseInsensitiveDict(pywr_resp.headers),
        )
        resp.session = self.session
        await pywr_resp.dispose()
        return resp

    def get(self, url: str, **kwargs):
        return self.request('GET', url, **kwargs)

    def patch(self, url: str, **kwargs):
        return self.request('PATCH', url, **kwargs)

    def put(self, url: str, **kwargs):
        return self.request('PUT', url, **kwargs)

    def post(self, url: str, **kwargs):
        return self.request('POST', url, **kwargs)

    def delete(self, url: str, **kwargs):
        return self.request('DELETE', url, **kwargs)

    def head(self, url: str, **kwargs):
        return self.request('HEAD', url, **kwargs)

    async def _setHeaders(self, headers: Union[dict, CaseInsensitiveDict]):
        '''
        Set the browser headers

        Parameters:
            headers (Union[dict, CaseInsensitiveDict]): Headers to set
        '''
        await self.context.set_extra_http_headers(
            {
                **headers,
                # convert lists to comma separated
                **{k: ', '.join(v) for k, v in headers.items() if isinstance(v, list)},
            }
        )

    async def _loadText(self, text):
        # load content into page
        await self.page.set_content(text)
        await self.page.wait_for_load_state('domcontentloaded')

    async def _setCookies(self, cookiejar: RequestsCookieJar):
        # convert cookiejar to list of dicts
        cookie_renders = cookiejar_to_list(cookiejar)
        # set cookies in playwright instance
        await self.context.add_cookies(cookie_renders)

    async def _close(self):
        if self._closed:
            # browser was closed, nothing to do
            return
        cookiejar = await self._getCookies()
        # update session if provided
        if self.session:
            self.session.cookies = cookiejar
        # update response
        if self.resp is not None:
            self.resp.cookies = cookiejar
            self.resp._content = await self.page.content()
            self.resp._text = None
            self.resp.url = self.page.url
            self.resp.status_code = self.status_code
        # close browser
        await self.shutdown()

    def __del__(self):
        self.close()


def render(
    url: str = None,
    headless: bool = True,
    proxy: dict = None,
    response: playfulrequests.response.Response = None,
    session: playfulrequests.session.TLSSession = None,
    mock_human: bool = False,
    extensions: Optional[Union[str, Iterable[str]]] = None,
    browser: Optional[Literal['firefox', 'chrome']] = None,
):
    assert any(
        (url, session, response is not None)
    ), 'Must provide a url or an existing session, response'

    if proxy:
        proxy = list(proxy.values())[0]
    render_session = BrowserSession(
        session=session,
        resp=response,
        proxy_ip=proxy,
        headless=headless,
        mock_human=mock_human,
        extensions=extensions,
        browser=browser,
    )
    # include headers from session if a TLSSession is provided
    if session and isinstance(session, playfulrequests.session.TLSSession):
        render_session.setHeaders(session.headers)
    # include merged cookies from session or from response
    if req_src := session or response:
        render_session.setCookies(req_src.cookies)
    if url:
        # goto url if url was provided
        render_session.goto(url)
    else:
        # load response content if a response was provided
        render_session.loadText(response.text)
    return render_session
